@@ -143,9 +143,9 @@ struct help_type help_table[] =
 		"BREAK",
 		"<178>Command<078>: #break\n"
 		"\n"
-		"         The break command can be used inside the #FOREACH, #LOOP, #PARSE, #WHILE and\n"
-		"         #SWITCH commands. When #BREAK is found, tintin will stop executing the command\n"
-		"         it is currently in and move on to the next.\n"
+		"         The break command can be used inside the #FOREACH, #LOOP, #PARSE,\n"
+		"         #WHILE and #SWITCH statements. When #BREAK is found, tintin will stop\n"
+		"         executing the statement it is currently in and move on to the next.\n"
 		"\n"
 		"<178>Example<078>: #while {1} {#math cnt $cnt + 1;#if {$cnt == 20} {#break}}\n"
 	},
@@ -257,8 +257,14 @@ struct help_type help_table[] =
 		"         foreground colors use <<888>g00> to <<888>g23>, for grayscale background colors\n"
 		"         use <<888>G00> to <<888>G23>.\n"
 		"\n"
-		"<178>Example<078>: #showme <<888>115>t<<888>828>e<<888>838>s<<888>848>t<<888>088> <<888>468>!<<888>568>!<<888>088>\n"
-		"         This will show: <115>t<828>e<838>s<848>t<088> <468>!<568>!<088>.\n"
+		"         The tertiary colors are as follows:\n"
+		"\n"
+		"         <<888>acf> - Azure            <<888>afc> - Viridian\n"
+		"         <<888>caf> - Indigo           <<888>cfa> - Lime\n"
+		"         <<888>fac> - Pink             <<888>fca> - Orange\n"
+		"\n"
+		"<178>Example<078>: #showme <<888>acf>Azure    <<888>afc>Viridian <<888>caf>Indigo\n"
+		"<178>Example<078>: #showme <<888>cfa>Lime     <<888>fac>Pink     <<888>fca>Orange\n"
 	},
 	{
 		"CONFIG",
@@ -285,9 +291,10 @@ struct help_type help_table[] =
 		"CONTINUE",
 		"<178>Command<078>: #continue\n"
 		"\n"
-		"         The continue command can be used inside the #FOREACH, #LOOP, #PARSE, #WHILE\n"
-		"         and #SWITCH commands. When #CONTINUE is found, tintin will go to the end\n"
-		"         of the command and proceed as normal, which may be to reiterate the command.\n"
+		"         The continue command can be used inside the #FOREACH, #LOOP, #PARSE,\n"
+		"         #WHILE and #SWITCH commands. When #CONTINUE is found, tintin will go\n"
+		"         to the end of the command and proceed as normal, which may be to\n"
+		"         reiterate the command.\n"
 		"\n"
 		"<178>Example<078>: #loop 1 10 cnt {#if {$cnt % 2 == 0} {#continue} {say $cnt}}\n"
 	},
@@ -414,23 +421,41 @@ struct help_type help_table[] =
 		"         a brief description. Use #event %* to see the current list of available\n"
 		"         events you can use.\n"
 		"\n"
+		"         DATE                 %1 month - %3 day   %4 hour : %5 minute\n"
+		"         DAY                  %3 day\n"
 		"         END OF PATH\n"
+		"         HOUR                 %4 hour\n"
 		"         IAC <VAR> <VAR>\n"
+		"         IAC SB GMCP <MODULE> %0 data     %1 raw data\n"
 		"         IAC SB MSSP          %0 variable %1 value\n"
 		"         IAC SB MSDP          %0 variable %1 value\n"
+		"         IAC SB MSDP <VAR>    %1 value\n"
 		"         IAC SB NEW-ENVIRON   %0 variable %1 value\n"
 		"         IAC SB ZMP <VAR>     %0 value\n"
 		"         IAC SB <VAR>         %0 raw text %1 raw data\n"
-		"         MAP ENTER ROOM       %0 vnum\n"
-		"         MAP EXIT ROOM        %0 vnum\n"
+		"         MAP ENTER ROOM       %0 new vnum %1 old vnum\n"
+		"         MAP ENTER ROOM <VAR> %0 new vnum %1 old vnum\n"
+		"         MAP EXIT ROOM        %0 old vnum %1 new vnum\n"
+		"         MAP EXIT ROOM <VAR>  %0 old vnum %1 new vnum\n"
+		"         MINUTE               %5 minute\n"
+		"         MONTH                %1 month\n"
+		"         PROGRAM START        %0 client name %1 client version\n"
 		"         PROGRAM TERMINATION\n"
 		"         RECEIVED INPUT       %0 raw text\n"
 		"         RECEIVED LINE        %0 raw text %1 plain text\n"
-		"         SEND OUTPUT          %0 raw output\n"
+		"         RECEIVED OUTPUT      %0 raw text\n"
+		"         SCREEN RESIZE        %0 cols %1 rows\n"
+		"         SECOND               %6 second\n"
+		"         SEND OUTPUT          %0 raw text\n"
+		"         SESSION ACTIVATED    %0 name\n"
 		"         SESSION CONNECTED    %0 name %1 host %2 ip %3 port\n"
+		"         SESSION DEACTIVATED  %0 name\n"
 		"         SESSION DISCONNECTED %0 name %1 host %2 ip %3 port\n"
+		"         TIME                 %4 hour : %5 minute : %6 second\n"
+		"         WEEK                 %2 week\n"
+		"         YEAR                 %0 year\n"
 		"\n"
-		"<178>Example<078>: #event {session connected} {#read mychar.tin}\n"
+		"<178>Example<078>: #event {SESSION CONNECTED} {#read mychar.tin}\n"
 		"\n"
 		"<178>Comment<078>: You can remove an event with the #unevent command.\n"
 	},
@@ -470,6 +495,7 @@ struct help_type help_table[] =
 		"         #format {test} {%a}   {number}  print corresponding ascii character\n"
 		"         #format {test} {%c}   {string}  use a highlight color name\n"
 		"         #format {test} {%d}   {number}  print a number with integer formatting\n"
+		"         #format {test} {%g}   {number}  perform thousand grouping on {number}\n"
 		"         #format {test} {%h}   {string}  turn text into a header line\n"
 		"         #format {test} {%l}   {string}  lowercase text\n"
 		"         #format {test} {%m}   {string}  perform mathematical calculation\n"
@@ -480,8 +506,8 @@ struct help_type help_table[] =
 		"         #format {test} {%t}   {format}  display time with strftime format\n"
 		"         #format {test} {%u}   {string}  uppercase text\n"
 		"         #format {test} {%w}   {string}  wordwrap text\n"
+		"         #format {test} {%A}     {char}  print corresponding ascii value\n"
 		"         #format {cols} {%C}         {}  store the screen width in {cols}\n"
-		"         #format {test} {%G}   {number}  perform thousand grouping on {number}\n"
 		"         #format {test} {%L}   {string}  store the string length in {test}\n"
 		"         #format {rows} {%R}         {}  store the screen height in {rows}\n"
 		"         #format {time} {%T}         {}  store the epoch time in {time}\n"
@@ -495,7 +521,8 @@ struct help_type help_table[] =
 		"         replace the function call with the line of text generated by the\n"
 		"         function.\n"
 		"\n"
-		"         Be aware that each function should set the $result variable.\n"
+		"         Be aware that each function should set the $result variable at the\n"
+		"         end of the function, or call #return with the given result.\n"
 		"\n"
 		"         To use a function use the @ character before the function name.\n"
 		"         The function arguments should be placed between braces behind the\n"
@@ -504,8 +531,6 @@ struct help_type help_table[] =
 		"         The function itself can use the provided arguments which are stored\n"
 		"         in %1 to %9, with %0 holding all arguments.\n"
 		"\n"
-		"         Currently you can't use a function from within another function\n"
-		"\n"
 		"<178>Example<078>: #function {rnd} {#math {result} {1 d (%2 - %1 + 1) + %1 - 1}}\n"
 		"         #showme A random number between 100 and 200: @rnd{100 200}\n"
 		"\n"
@@ -524,20 +549,22 @@ struct help_type help_table[] =
 		"\n"
 		"<178>Comment<078>: You can remove a gag with the #ungag command.\n"
 	},
+
 	{
 		"GREETING",
-		"<068>      #<068>###################################################################<068>#\n"
-		"<068>      #<078>                                                                   <068>#\n"
-		"<068>      #<078>                     T I N T I N + +   "VERSION_NUM"                      <068>#\n"
-		"<068>      #<078>                                                                   <068>#\n"
-		"<068>      #<078>           (<068>T<078>)he k(<068>I<078>)cki(<068>N<078>) (<068>T<078>)ickin d(<068>I<078>)kumud clie(<068>N<078>)t <068>           #\n"
-		"<068>      #<078>                                                                   <068>#\n"
-		"<068>      #<078>         Code by Peter Unold, Bill Reis, David A. Wagner,          <068>#\n"
-		"<068>      #<078>      Rob Ellsworth, Jeremy C. Jack, and Igor van den Hoven.       <068>#\n"
-		"<068>      #<078>                                                                   <068>#\n"
-//		"<068>      #<078>                             1992, 2010                            <068>#\n"
-		"<068>      #<068>###################################################################<068>#<088>\n"
+		"<068>      #<068>##################################################################<068>#\n"
+		"<068>      #<078>                                                                  <068>#\n"
+		"<068>      #<078>                     T I N T I N + +   "CLIENT_VERSION"                   <068>#\n"
+		"<068>      #<078>                                                                  <068>#\n"
+		"<068>      #<078>           (<068>T<078>)he k(<068>I<078>)cki(<068>N<078>) (<068>T<078>)ickin d(<068>I<078>)kumud clie(<068>N<078>)t <068>          #\n"
+		"<068>      #<078>                                                                  <068>#\n"
+		"<068>      #<078>         Code by Peter Unold, Bill Reis, David A. Wagner,         <068>#\n"
+		"<068>      #<078>      Rob Ellsworth, Jeremy C. Jack, and Igor van den Hoven.      <068>#\n"
+		"<068>      #<078>                                                                  <068>#\n"
+		"<068>      #<078>                             1992, 2012                           <068>#\n"
+		"<068>      #<068>##################################################################<068>#<088>\n"
 	},
+
 	{
 		"GREP",
 		"<178>Command<078>: #grep <178>[<078>page<178>]<078> <178>{<078>search string<178>}<078>\n"
@@ -548,6 +575,9 @@ struct help_type help_table[] =
 		"         use wildcards for better search results. Be aware the search string\n"
 		"         is case sensitive, which can be disabled by using %i.\n"
 		"\n"
+		"         By default grep searches from the end of the scrollback buffer to the\n"
+		"         beginning, this can be reversed by using a negative page number.\n"
+		"\n"
 		"<178>Example<078>: #grep Bubba tells you\n"
 		"         This will show all occasions where bubba tells you something.\n"
 	},
@@ -564,12 +594,18 @@ struct help_type help_table[] =
 		"<178>Command<078>: #highlight <178>{<078>string<178>}<078> <178>{<078>color names<178>}<078>\n"
 		"\n"
 		"         The highlight command is used to allow you to highlight strings of text\n"
-		"         from the mud.  Available color names are:\n"
+		"         from the mud.  Available ANSI color names are:\n"
 		"\n"
 		"         reset, light, faint, underscore, blink, reverse, dim,\n"
 		"         black, red, green, yellow, blue, magenta, cyan, white,\n"
 		"         b black, b red, b green, b yellow, b blue, b magenta, b cyan, b white\n"
 		"\n"
+		"         Available xterm 256 color names are:\n"
+		"\n"
+		"         dark azure, dark ebony, dark jade, dark lime, dark orange, dark silver,\n"
+		"         dark tan, dark violet, light azure, light ebony, light jade, light lime,\n"
+		"         light orange, light silver, light tan, light violet.\n"
+		"\n"
 		"         The %1-99 variables can be used as 'wildcards' that will match with any\n"
 		"         text. They are useful for highlighting a complete line. The %0 variable\n"
 		"         should never be used in highlights.\n"
@@ -665,9 +701,22 @@ struct help_type help_table[] =
 		"\n"
 		"         #line log {filename} {[text]}          Log the current or given line to\n"
 		"                                                file.\n"
+		"\n"
 		"         #line logverbatim {filename} {[text]}  Like log with no substitutions.\n"
+		"\n"
 		"         #line gag                              Gag the next line.\n"
 		"\n"
+		"         #line ignore {argument}                Argument is executed without\n"
+		"                                                any triggers being checked.\n"
+		"\n"
+		"         #line substitute {options} {argument}  Substitutes the given options:\n"
+		"                                                variables, functions, colors,\n"
+		"                                                escapes, secure, eol, lnf, in\n"
+		"                                                the given argument next executes\n"
+		"                                                it as a command.\n"
+		"\n"
+		"         #line verbose {argument}               Argument is executed verbose.\n"
+		"\n"
 		"         When using #line log and logging in html format use \\c< \\c> \\c& \\c\" to\n"
 		"         log a literal < > & and \".\n"
 	},
@@ -694,6 +743,9 @@ struct help_type help_table[] =
 		"         When inserting an item a positive index will prepend the item at the\n"
 		"         given index, while a negative index will append the item.\n"
 		"\n"
+		"         The add and create options allow using multiple items, as well\n"
+		"         as semi-colon separated items.\n"
+		"\n"
 		"         A length of 0 is returned for an empty or non existant list.\n"
 		"\n"
 		"         You can directly access elements in a list variable using $var[1],\n"
@@ -736,8 +788,12 @@ struct help_type help_table[] =
 		"         Another option is pressing ctrl-v, which will enable CONVERT META for\n"
 		"         the next key pressed.\n"
 		"\n"
-		"<178>Example<078>: #macro {(press ctrl-v)(press F1)} {shout tintin is the greatest!\n"
-		"         The next time you press F1 the entered command should be executed.\n"
+		"<178>Example<078>: #macro {(press ctrl-v)(press F1)} {#showme \\e[2J;#buffer lock}\n"
+		"         Clear the screen and lock the window when you press F1, useful when the\n"
+		"         boss is near.\n"
+		"\n"
+		"<178>Example<078>: #macro {\\eOM} {#cursor enter}\n"
+		"         Makes the keypad's enter key work as an enter in keypad mode.\n"
 		"\n"
 		"<178>Comment<078>: Not all terminals properly initialize the keypad key sequences.\n"
 		"         If this is the case you can still use the keypad, but instead of the\n"
@@ -753,7 +809,9 @@ struct help_type help_table[] =
 		"         to use mapping for muds using 'ne' 'se' 'sw' 'nw' directions, you need\n"
 		"         to create a #pathdir for each direction.\n"
 		"\n"
-		"         #map color: Sets the map color of the current room you are in.\n"
+		"         #map at <location> <command>: Execute the command at the location.\n"
+		"\n"
+		"         #map color <field> value: Sets the map color for the given field.\n"
 		"\n"
 		"         #map create <size>: Creates a new map and room 1, required to start\n"
 		"                  mapping. Default size is 15000 rooms.\n"
@@ -762,21 +820,23 @@ struct help_type help_table[] =
 		"\n"
 		"         #map delete <direction>: Deletes the room in the given direction.\n"
 		"\n"
-		"         #map dig <direction> [new]: Creates an exit in the given direction. If\n"
-		"                  no valid direction is given or no existing room is found a new\n"
-		"                  room is created. Useful for portal links and other alternative\n"
-		"                  forms of transportation. If the 'new' argument is provided all\n"
-		"                  existing rooms are ignored.\n"
+		"         #map dig <direction|vnum> [new|<vnum>]: Creates an exit in the given\n"
+		"                  direction. If no valid direction is given or no existing room\n"
+		"                  is found a new room is created. Useful for portal links and\n"
+		"                  other alternative forms of transportation. If the 'new'\n"
+		"                  argument is provided all existing rooms are ignored. If a\n"
+		"                  room vnum is given as the second argument an exit will be\n"
+		"                  created to the given room vnum. If the direction argument is\n"
+		"                  a valid unused vnum that room is created.\n"
 		"\n"
-		"         #map exit <exit name> <exit commands>: Sets the exit command. This is\n"
-		"                  useful if you are dealing with a closed door where you would\n"
-		"                  set the exit command with: '#map exit e {open east;e}'.\n"
-		"                  Keep in mind the exit command is only used when using\n"
-		"                  '#map find' and '#map walk'.\n"
+		"         #map exit <exit> <COMMAND|DIRECTION|FLAG|GET|NAME|SET|VNUM> <argument>\n"
+		"                  Set the exit data. Useful with a closed door where you can\n"
+		"                  set the exit command: '#map exit e command {open east;e}'.\n"
 		"\n"
-		"         #map exitdir <exit name> <exit direction>: Set the exit direction. This\n"
-		"                  is useful if you have an unusual type of exit with a known\n"
-		"                  direction, like for example a tree that can be climbed.\n"
+		"         #map exitflag hide: When set, '#map' will not display the map beyond\n"
+		"                  this exit. When mapping overlapping areas or areas that aren't\n"
+		"                  build consistently you need this flag as well to stop\n"
+		"                  auto-linking, unless you use void rooms.\n"
 		"\n"
 		"         #map explore <direction>: Follows the direction until a dead end or an\n"
 		"                  intersection is found. The route is stored in #path and can\n"
@@ -792,26 +852,35 @@ struct help_type help_table[] =
 		"                  found. These options are also available to the goto, and link\n"
 		"                  commands.\n"
 		"\n"
-		"         #map flag static: Will make the map static so new rooms are no longer\n"
-		"                  created when walking into an unmapped direction. Useful when\n"
-		"                  you're done mapping and regularly bump into walls accidentally\n"
-		"                  creating a new room. #map dig etc will still work.\n"
-		"\n"
-		"         #map flag vtmap: Will enable the vtmap which is shown in the top split\n"
-		"                  screen if you have one. You can create a 10 rows high top\n"
-		"                  screen by using '#split 10 1'.\n"
-		"\n"
 		"         #map flag asciigraphics: Draws a smaller but more detailed map that\n"
 		"                  displays the ne se sw nw exits and room symbols.\n"
 		"\n"
 		"         #map flag asciivnums: Display room vnums with asciigraphics enabled.\n"
 		"\n"
+		"         #map flag nofollow: When you enter movement commands the map will no\n"
+		"                  longer automatically follow along. Useful for MSDP and GMCP\n"
+		"                  automapping scripts.\n"
+		"\n"
+		"         #map flag static: Will make the map static so new rooms are no longer\n"
+		"                  created when walking into an unmapped direction. Useful when\n"
+		"                  you're done mapping and regularly bump into walls accidentally\n"
+		"                  creating a new room. #map dig etc will still work.\n"
+		"\n"
 		"         #map flag vtgraphics: Enables vt line drawing on some terminals\n"
 		"\n"
+		"         #map flag vtmap: Will enable the vtmap which is shown in the top split\n"
+		"                  screen if you have one. You can create a 16 rows high top\n"
+		"                  screen by using '#split 16 1'.\n"
+		"\n"
+		"         #map goto <room vnum> [dig]: Takes you to the given room vnum, with the\n"
+		"                  dig argument a new room will be created if non exists.\n"
+		"\n"
 		"         #map goto <room name> [exits]: Takes you to the given room name, if\n"
 		"                  you provide exits those must match as well.\n"
 		"\n"
-		"         #map get <option> <variable>: Store a map value into a variable.\n"
+		"         #map get <option> <variable> [vnum]: Store a map value into a\n"
+		"                    variable, if no vnum is given the current room is used.\n"
+		"\n"
 		"         #map get roomexits <variable>: Store all room exits into variable.\n"
 		"\n"
 		"         #map info: Gives information about the map and room you are in.\n"
@@ -819,7 +888,7 @@ struct help_type help_table[] =
 		"         #map insert <direction> [roomflag]: Insert a room in the given\n"
 		"                  direction. Most useful for inserting void rooms.\n"
 		"\n"
-		"         #map jump <x> <y>: Jump to the given coordinate, which is relative\n"
+		"         #map jump <x> <y> <z>: Jump to the given coordinate, which is relative\n"
 		"                  to your current room.\n"
 		"\n"
 		"         #map leave: Makes you leave the map. Useful when entering a maze.\n"
@@ -835,14 +904,17 @@ struct help_type help_table[] =
 		"                  font for drawing characters use:\n"
 		"                  #loop 32 255 cnt {#echo {%-3s  %a} $cnt $cnt}\n"
 		"\n"
-		"         #map link <direction> <room name>: Links two rooms. If a valid\n"
-		"                  direction is given the link is two way.\n"
+		"         #map link <direction> <room name> [both]: Links two rooms. If the both\n"
+		"                  argument and a valid direction is given the link is two way.\n"
 		"\n"
-		"         #map list <location>: Lists all matching rooms and their distance.\n"
+		"         #map list <roomname> <exits> <roomdesc> <roomarea> <roomnote> <terrain>\n"
+		"                  Lists all matching rooms and their distance.\n"
 		"\n"
-		"         #map map {<x>x<y>} {filename}: shows a map of surrounding rooms. The\n"
-		"                  {horizontal x vertical} argument i.e 80x25 is optional, and so\n"
-		"                  is the filename argument to log the map output to file.\n"
+		"         #map map {<x>x<y>} {filename} {a}: shows a map of surrounding rooms.\n"
+		"                  The {horizontal x vertical} argument i.e 80x25 is optional,\n"
+		"                  and so is the filename argument to log the map output to\n"
+		"                  file. The {a} argument causes data to be appended so it can\n"
+		"                  easily be tailed with the tail -f <file> shell command.\n"
 		"\n"
 		"         #map move <direction>: This does the same as an actual movement\n"
 		"                  command, updating your location on the map and creating new\n"
@@ -855,6 +927,9 @@ struct help_type help_table[] =
 		"\n"
 		"         #map read <filename>: Will load the given map file.\n"
 		"\n"
+		"         #map resize <size>: Resize the map, setting the maximum number of\n"
+		"                  rooms.\n"
+		"\n"
 		"         #map return: Returns you to your last known room after leaving the map\n"
 		"                  or loading a map.\n"
 		"\n"
@@ -863,7 +938,7 @@ struct help_type help_table[] =
 		"         #map roomflag hide: When set, '#map' will not display the map beyond\n"
 		"                  this room. When mapping overlapping areas or areas that aren't\n"
 		"                  build consistently you need this flag as well to stop\n"
-		"                  auto-linking.\n"
+		"                  auto-linking, unless you use void rooms.\n"
 		"         #map roomflag leave: When entering a room with this flag, you will\n"
 		"                  automatically leave the map. Useful when set at the entrance\n"
 		"                  of an unmappable maze.\n"
@@ -879,7 +954,8 @@ struct help_type help_table[] =
 		"                  requires using braces. Besides the room name a list of\n"
 		"                  exits can be provided for more precise matching.\n"
 		"\n"
-		"         #map set <option> <value>: Set a map value for your current room.\n"
+		"         #map set <option> <value> [vnum]: Set a map value for your current\n"
+		"                  room, or given room if a room vnum is provided.\n"
 		"\n"
 		"         #map travel <direction> <delay>: Follows the direction until a dead end\n"
 		"                  or an intersection is found. Use braces around the direction\n"
@@ -890,11 +966,16 @@ struct help_type help_table[] =
 		"                  they will be deleted, otherwise you'll simply move back a\n"
 		"                  room. Useful if you walked into a non existant direction.\n"
 		"\n"
+		"         #map uninsert <direction>: Exact opposite of the insert command.\n"
+		"\n"
 		"         #map unlink <direction> [both]: Will remove the exit, this isn't two\n"
 		"                  way so you can have the map properly display no exit rooms and\n"
 		"                  mazes.\n"
 		"                  If you use the both argument the exit is removed two-ways.\n"
-
+		"\n"
+		"         #map vnum <low> [high]: Change the room vnum to the given number, if\n"
+		"                  a range is provided the first available room in that range\n"
+		"                  is selected.\n"
 		"\n"
 		"         #map write <filename>: Will save the map.\n"
 	},
@@ -989,18 +1070,20 @@ struct help_type help_table[] =
 		"PATH",
 		"<178>Command<078>: #path <178>{<078>del<178>|<078>end<178>|<078>ins<178>|<078>load<178>|<078>map<178>|<078>new<178>|<078>run<178>|<078>save<178>|<078>walk<178>}<078> <178>{<078>argument<178>}<078>\n"
 		"\n"
-		"         #path del:  Will delete the last move of the path.\n"
-		"         #path end:  Movement commands are no longer added to the path.\n"
-		"         #path ins:  Add the given argument to the path.\n"
-		"         #path load: Load the given variable as the new path.\n"
-		"         #path map:  Display the so far created path.\n"
-		"         #path new:  Movement commands will be added to the path.\n"
-		"         #path run:  Enter all the steps of the current path, with\n"
-		"                     an optional given delay.\n"
-		"         #path save: Save the created path to a variable. You must specify\n"
-		"                     whether you want to save the path forward or backward.\n"
-		"         #path walk: Take one step forward of the current path. You can walk\n"
-		"                     backwards using #path walk backwards.\n"
+		"         #path del:   Will delete the last move of the path.\n"
+		"         #path end:   Movement commands are no longer added to the path.\n"
+		"         #path ins:   Add the given argument to the path.\n"
+		"         #path load:  Load the given variable as the new path.\n"
+		"         #path new:   Movement commands will be added to the path.\n"
+		"         #path run:   Enter all the steps of the current path, with\n"
+		"                      an optional given delay.\n"
+		"         #path save:  Save the created path to a variable. You must specify\n"
+		"                      whether you want to save the path forward or backward.\n"
+		"         #path show:  Display the so far created path.\n"
+		"         #path unzip: Load the given speedwalk as the new path.\n"
+		"         #path walk:  Take one step forward of the current path. You can\n"
+		"                      walk backwards using #path walk backwards.\n"
+		"         #path zip:   Turn loaded path into a speedwalk.\n"
 		"\n"
 		"<178>Example<078>: #path ins {unlock n;open n} {unlock s;open s}\n"
 	},
@@ -1355,6 +1438,8 @@ struct help_type help_table[] =
 		"\n"
 		"         Executes given command every # of seconds.\n"
 		"\n"
+		"<178>Comment<078>: Tickers don't work in the startup session.\n"
+		"\n"
 		"<178>Comment<078>: You can remove a ticker with the #unticker command.\n"
 	},
 	{
@@ -1394,6 +1479,27 @@ struct help_type help_table[] =
 		"\n"
 		"<178>Comment<078>: You can remove a variable with the #unvariable command.\n"
 	},
+	{
+		"WHILE",
+		"<178>Command<078>: #while <178>{<078>conditional<178>} {<078>commands<178>}<078>\n"
+		"\n"
+		"         This command works similar to a 'while' statement in other languages.\n"
+		"\n"
+		"         When a 'while' command is encourated, the conditional is evaluated,\n"
+		"         and if TRUE (any non-zero result) the commands are executed. The\n"
+		"         'while' loop will be repeated indefinitely until the conditional is\n"
+		"         FALSE or the #BREAK or #RETURN commands are found.\n"
+		"\n"
+		"         The 'while' statement is only evaluated if it is read, so you must\n"
+		"         nest it inside a trigger, like an alias or action.\n"
+		"\n"
+		"         The conditional is evaluated exactly the same as in the 'math' command.\n"
+                "\n"
+                "<178>Example<078>: #math cnt 0;#while {$cnt < 20} {#math cnt $cnt + 1;say $cnt}\n"
+		"\n"
+		"<178>Comment<078>: See '#help math', for more information.\n"
+	},
+                                                                                                   
 	{
 		"WRITE",
 		"<178>Command<078>: #write <178>{<078>filename<178>}<078>\n"
@@ -1403,10 +1509,11 @@ struct help_type help_table[] =
 	},
 	{
 		"ZAP",
-		"<178>Command<078>: #zap\n"
+		"<178>Command<078>: #zap {[session]}\n"
 		"\n"
 		"         Kill your current session.  If there is no current session, it will\n"
-		"         cause the program to terminate.\n"
+		"         cause the program to terminate. If you provide an argument it'll zap\n"
+		"         the given session instead.\n"
 	},
 	{
 		"",
@@ -1420,7 +1527,7 @@ DO_COMMAND(do_help)
 	char left[BUFFER_SIZE], add[BUFFER_SIZE], buf[BUFFER_SIZE], *ptf, *pto;
 	int cnt, found;
 
-	arg = get_arg_in_braces(arg, left, TRUE);
+	arg = get_arg_in_braces(ses, arg, left, TRUE);
 
 	if (*left == 0)
 	{
@@ -1441,7 +1548,7 @@ DO_COMMAND(do_help)
 
 		for (cnt = 0 ; *help_table[cnt].name != 0 ; cnt++)
 		{
-			if (is_abbrev(left, help_table[cnt].name) || atoi(left) == cnt + 1 || match(ses, help_table[cnt].name, left))
+			if (is_abbrev(left, help_table[cnt].name) || atoi(left) == cnt + 1 || match(ses, help_table[cnt].name, left, SUB_VAR|SUB_FUN))
 			{
 				substitute(ses, help_table[cnt].text, buf, SUB_COL);
 
