@@ -28,6 +28,7 @@
 #include "tintin.h"
 
 #include <errno.h>
+#include <signal.h>
 
 /************************/
 /* the #session command */
@@ -43,7 +44,7 @@ DO_COMMAND(do_session)
 
 	arg = temp;
 
-	arg = get_arg_in_braces(arg, left,  FALSE);
+	arg = get_arg_in_braces(ses, arg, left,  FALSE);
 
 	if (*left == 0)
 	{
@@ -95,7 +96,7 @@ void show_session(struct session *ses, struct session *ptr)
 {
 	char temp[BUFFER_SIZE];
 
-	sprintf(temp, "%-12s%20s:%-5s", ptr->name, ptr->host, ptr->port);
+	sprintf(temp, "%3d %-12s %20s:%-5s", ptr->socket, ptr->name, ptr->host, ptr->port);
 
 	if (ptr == gtd->ses)
 	{
@@ -138,12 +139,7 @@ struct session *newactive_session(void)
 	}
 	else
 	{
-		gtd->ses = gts;
-		dirty_screen(gtd->ses);
-
-		tintin_printf(NULL, "");
-
-		tintin_printf(NULL, "#THERE'S NO ACTIVE SESSION NOW.");
+		activate_session(gts);
 	}
 	pop_call();
 	return gtd->ses;
@@ -151,12 +147,16 @@ struct session *newactive_session(void)
 
 struct session *activate_session(struct session *ses)
 {
+	check_all_events(gtd->ses, SUB_ARG|SUB_SEC, 0, 1, "SESSION DEACTIVATED", gtd->ses->name);
+
 	gtd->ses = ses;
 
 	dirty_screen(ses);
 
 	tintin_printf(ses, "#SESSION '%s' ACTIVATED.", ses->name);
 
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "SESSION ACTIVATED", ses->name);
+
 	return ses;
 }
 
@@ -178,8 +178,8 @@ struct session *new_session(struct session *ses, char *name, char *address, int
 		return ses;
 	}
 
-	address = get_arg_in_braces(address, host, FALSE);
-	address = get_arg_in_braces(address, port, FALSE);
+	address = get_arg_in_braces(ses, address, host, FALSE);
+	address = get_arg_in_braces(ses, address, port, FALSE);
 
 	if (desc == 0)
 	{
@@ -235,16 +235,26 @@ struct session *new_session(struct session *ses, char *name, char *address, int
 		newsession->list[cnt] = copy_list(newsession, gts->list[cnt], cnt);
 	}
 
-	init_screen_size(newsession);
+	newsession->rows          = gts->rows;
+	newsession->cols          = gts->cols;
+	newsession->top_row       = gts->top_row;
+	newsession->bot_row       = gts->bot_row;
 
 	init_buffer(newsession, gts->scroll_max);
 
-	dirty_screen(newsession);
-
-	tintin_printf2(ses, "#Trying to connect to %s port %s.", newsession->host, newsession->port);
+	if (desc)
+	{
+		tintin_printf(ses, "#TRYING TO LAUNCH '%s' RUNNING '%s'.", newsession->name, newsession->host);
+	}
+	else
+	{
+		tintin_printf(ses, "#TRYING TO CONNECT '%s' TO '%s' PORT '%s'.", newsession->name, newsession->host, newsession->port);
+	}
 
 	gtd->ses = newsession;
 
+	dirty_screen(newsession);
+
 	if (desc == 0)
 	{
 		connect_session(newsession);
@@ -295,12 +305,7 @@ void connect_session(struct session *ses)
 
 		tintin_printf(ses, "#SESSION '%s' CONNECTED TO '%s' PORT '%s'", ses->name, ses->host, ses->port);
 
-		if (atoi(ses->port) == TELNET_PORT)
-		{
-			init_telnet_session(ses);
-		}
-
-		check_all_events(ses, 0, 4, "SESSION CONNECTED", ses->name, ses->host, ses->ip, ses->port);
+		check_all_events(ses, SUB_ARG|SUB_SEC, 0, 4, "SESSION CONNECTED", ses->name, ses->host, ses->ip, ses->port);
 
 		return;
 	}
@@ -362,11 +367,17 @@ void cleanup_session(struct session *ses)
 		{
 			syserr("close in cleanup");
 		}
+
+		// the PID is stored in the session's port.
+
+		if (HAS_BIT(ses->flags, SES_FLAG_RUN))
+		{
+			kill(atoi(ses->port), SIGKILL);
+		}
+
 	}
 
-	check_all_events(ses, 0, 4, "SESSION DISCONNECTED", ses->name, ses->host, ses->ip, ses->port);
-
-	tintin_printf(gtd->ses, "");
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 4, "SESSION DISCONNECTED", ses->name, ses->host, ses->ip, ses->port);
 
 	tintin_printf(gtd->ses, "#SESSION '%s' DIED.", ses->name);
 
@@ -409,6 +420,12 @@ void dispose_session(struct session *ses)
 		delete_map(ses);
 	}
 
+	if (ses->mccp)
+	{
+		inflateEnd(ses->mccp);
+		free(ses->mccp);
+	}
+
 	init_buffer(ses, 0);
 
 	free(ses->name);
