@@ -32,7 +32,7 @@
 struct command_type command_table[] =
 {
 	{    "action",            do_action,            TOKEN_TYPE_COMMAND },
-	{    "advertise",         do_advertise,         TOKEN_TYPE_COMMAND },
+//	{    "advertise",         do_advertise,         TOKEN_TYPE_COMMAND },
 	{    "alias",             do_alias,             TOKEN_TYPE_COMMAND },
 	{    "all",               do_all,               TOKEN_TYPE_COMMAND },
 	{    "bell",              do_bell,              TOKEN_TYPE_COMMAND },
@@ -93,7 +93,6 @@ struct command_type command_table[] =
 	{    "snoop",             do_snoop,             TOKEN_TYPE_COMMAND },
 	{    "split",             do_split,             TOKEN_TYPE_COMMAND },
 	{    "substitute",        do_substitute,        TOKEN_TYPE_COMMAND },
-	{    "suspend",           do_suspend,           TOKEN_TYPE_COMMAND },
 	{    "switch",            do_nop,               TOKEN_TYPE_SWITCH  },
 	{    "system",            do_system,            TOKEN_TYPE_COMMAND },
 	{    "tab",               do_tab,               TOKEN_TYPE_COMMAND },
@@ -146,6 +145,20 @@ struct list_type list_table[LIST_MAX] =
 	{    "VARIABLE",          "VARIABLES",          ALPHA,       2,  LIST_FLAG_SHOW|LIST_FLAG_MESSAGE|LIST_FLAG_READ|LIST_FLAG_WRITE|LIST_FLAG_CLASS|LIST_FLAG_INHERIT|LIST_FLAG_NEST }
 };
 
+struct substitution_type substitution_table[] =
+{
+//	{    "ARGUMENTS",            1  },
+	{    "VARIABLES",            2  },
+	{    "FUNCTIONS",            4  },
+	{    "COLORS",               8  },
+	{    "ESCAPES",             16  },
+//	{    "COMMANDS",            32  },
+	{    "SECURE",              64  },
+	{    "EOL",                128  },
+	{    "LNF",                256  },
+//	{    "FIX",               1024  },
+	{    "",                  0     }
+};
 
 struct config_type config_table[] =
 {
@@ -163,24 +176,31 @@ struct config_type config_table[] =
 		config_buffersize
 	},
 
+	{
+		"CHARSET",
+		"",
+		"The character set encoding used by TinTin++",
+		config_charset
+	},
+
 	{
 		"COLOR PATCH",
-		"Tintin will properly color the start of each line",
-		"Tintin will leave color handling to the server",
+		"TinTin++ will properly color the start of each line",
+		"TinTin++ will leave color handling to the server",
 		config_colorpatch
 	},
 
 	{
 		"CONNECT RETRY",
 		"",
-		"Seconds Tintin will try to connect before giving up",
+		"Seconds TinTin++ will try to connect before giving up",
 		config_connectretry
 	},
 
 	{
 		"CONVERT META",
-		"Tintin converts meta prefixed characters",
-		"Tintin doesn't convert meta prefixed characters",
+		"TinTin++ converts meta prefixed characters",
+		"TinTin++ doesn't convert meta prefixed characters",
 		config_convertmeta
 	},
 
@@ -205,6 +225,8 @@ struct config_type config_table[] =
 		config_commandecho
 	},
 
+
+
 	{
 		"HISTORY SIZE",
 		"",
@@ -221,8 +243,8 @@ struct config_type config_table[] =
 
 	{
 		"LOG LEVEL",
-		"Tintin only logs low level mud data",
-		"Tintin only logs high level mud data",
+		"TinTin++ only logs low level mud data",
+		"TinTin++ only logs high level mud data",
 		config_loglevel
 	},
 
@@ -236,7 +258,7 @@ struct config_type config_table[] =
 	{
 		"PACKET PATCH",
 		"",
-		"Seconds Tintin will try to patch broken packets",
+		"Seconds TinTin++ will try to patch broken packets",
 		config_packetpatch
 	},
 
@@ -271,14 +293,14 @@ struct config_type config_table[] =
 	{
 		"TINTIN CHAR",
 		"",
-		"The character used for Tintin commands",
+		"The character used for TinTin++ commands",
 		config_tintinchar
 	},
 
 	{
 		"VERBATIM",
-		"Your keyboard input isn't modified by Tintin",
-		"Your keyboard input is parsed by Tintin",
+		"Your keyboard input isn't modified by TinTin++",
+		"Your keyboard input is parsed by TinTin++",
 		config_verbatim
 	},
 
@@ -303,6 +325,13 @@ struct config_type config_table[] =
 		config_wordwrap
 	},
 
+	{
+		"256 COLORS",
+		"Your terminal is 256 color capable",
+		"Your terminal is not 256 color capable",
+		config_256color
+	},
+
 	{
 		"",
 		"",
@@ -313,32 +342,52 @@ struct config_type config_table[] =
 
 struct color_type color_table[] =
 {
-	{    "reset",          "0", "<088>" },
-	{    "light",          "1", "<188>" },
-	{    "bold",           "1", "<188>" },
-	{    "faint",          "2", "<288>" },
-	{    "underscore",     "4", "<488>" },
-	{    "blink",          "5", "<588>" },
-	{    "reverse",        "7", "<788>" },
-	{    "dim",           "22", "<288>" },
-	{    "dark",          "22", "<288>" },
-	{    "black",         "30", "<808>" },
-	{    "red",           "31", "<818>" },
-	{    "green",         "32", "<828>" },
-	{    "yellow",        "33", "<838>" },
-	{    "blue",          "34", "<848>" },
-	{    "magenta",       "35", "<858>" },
-	{    "cyan",          "36", "<868>" },
-	{    "white",         "37", "<878>" },
-	{    "b black",       "40", "<880>" },
-	{    "b red",         "41", "<881>" },
-	{    "b green",       "42", "<882>" },
-	{    "b yellow",      "43", "<883>" },
-	{    "b blue",        "44", "<884>" },
-	{    "b magenta",     "45", "<885>" },
-	{    "b cyan",        "46", "<886>" },
-	{    "b white",       "47", "<887>" },
-	{    "",              ""  , "<888>" }
+	{    "reset",         "<088>" },
+	{    "light",         "<188>" },
+	{    "bold",          "<188>" },
+	{    "faint",         "<288>" },
+	{    "dim",           "<288>" },
+	{    "dark",          "<288>" },
+	{    "underscore",    "<488>" },
+	{    "blink",         "<588>" },
+	{    "reverse",       "<788>" },
+
+	{    "black",         "<808>" },
+	{    "red",           "<818>" },
+	{    "green",         "<828>" },
+	{    "yellow",        "<838>" },
+	{    "blue",          "<848>" },
+	{    "magenta",       "<858>" },
+	{    "cyan",          "<868>" },
+	{    "white",         "<878>" },
+
+	{    "dark azure",    "<abd>" },
+	{    "dark ebony",    "<g04>" },
+	{    "dark jade",     "<adb>" },
+	{    "dark lime",     "<bda>" },
+	{    "dark orange",   "<dba>" },
+	{    "dark silver",   "<ccc>" },
+	{    "dark tan",      "<cba>" },
+	{    "dark violet",   "<bad>" },
+
+	{    "light azure",   "<acf>" },
+	{    "light ebony",   "<bbb>" },
+	{    "light jade",    "<afc>" },
+	{    "light lime",    "<cfa>" },
+	{    "light orange",  "<fca>" },
+	{    "light silver",  "<eee>" },
+	{    "light tan",     "<eda>" },
+	{    "light violet",  "<caf>" },
+
+	{    "b black",       "<880>" },
+	{    "b red",         "<881>" },
+	{    "b green",       "<882>" },
+	{    "b yellow",      "<883>" },
+	{    "b blue",        "<884>" },
+	{    "b magenta",     "<885>" },
+	{    "b cyan",        "<886>" },
+	{    "b white",       "<887>" },
+	{    "",              "<888>" }
 };
 
 struct class_type class_table[] =
@@ -409,36 +458,41 @@ struct array_type array_table[] =
 
 struct map_type map_table[] =
 {
+	{     "AT",               map_at,              1    },
 	{     "COLOR",            map_color,           1    },
 	{     "CREATE",           map_create,          0    },
+	{     "DEBUG",            map_debug,           2    },
 	{     "DELETE",           map_delete,          1    },
 	{     "DESTROY",          map_destroy,         1    },
 	{     "DIG",              map_dig,             2    },
-	{     "EXITCMD",          map_exitcmd,         2    },
-	{     "EXITDIR",          map_exitdir,         2    },
+	{     "EXIT",             map_exit,            2    },
+	{     "EXITFLAG",         map_exitflag,        2    },
 	{     "EXPLORE",          map_explore,         2    },
 	{     "FIND",             map_find,            1    },
-	{     "FLAG",             map_flag,            2    },
+	{     "FLAG",             map_flag,            1    },
 	{     "GET",              map_get,             2    },
 	{     "GOTO",             map_goto,            1    },
 	{     "INFO",             map_info,            1    },
 	{     "INSERT",           map_insert,          2    },
-	{     "JUMP",             map_jump,            1    },
+	{     "JUMP",             map_jump,            2    },
 	{     "LEAVE",            map_leave,           2    },
 	{     "LEGENDA",          map_legend,          1    },
 	{     "LINK",             map_link,            2    },
-	{     "LIST",             map_list,            1    },
+	{     "LIST",             map_list,            2    },
 	{     "MAP",              map_map,             2    },
 	{     "MOVE",             map_move,            2    },
 	{     "NAME",             map_name,            2    },
 	{     "READ",             map_read,            0    },
+	{     "RESIZE",           map_resize,          1    },
 	{     "RETURN",           map_return,          1    },
 	{     "ROOMFLAG",         map_roomflag,        2    },
 	{     "RUN",              map_run,             2    },
 	{     "SET",              map_set,             2    },
 	{     "TRAVEL",           map_travel,          2    },
 	{     "UNDO",             map_undo,            2    },
+	{     "UNINSERT",         map_uninsert,        2    },
 	{     "UNLINK",           map_unlink,          2    },
+	{     "VNUM",             map_vnum,            2    },
 	{     "WRITE",            map_write,           1    },
 	{     "",                 NULL,                0    }
 };
@@ -497,7 +551,7 @@ struct cursor_type cursor_table[] =
 		"CTRL DELETE",
 		"Delete one character, exit on an empty line",
 		"",
-		cursor_delete_exit
+		cursor_delete_or_exit
 	},
 	{
 		"DELETE",
@@ -506,11 +560,17 @@ struct cursor_type cursor_table[] =
 		cursor_delete
 	},
 	{
-		"DELETE WORD",
-		"Delete next word backward",
+		"DELETE WORD LEFT",
+		"Delete backwards till next space",
 		"",
 		cursor_delete_word_left
 	},
+	{
+		"DELETE WORD RIGHT",
+		"Delete forwards till next space",
+		"",
+		cursor_delete_word_right
+	},
 	{
 		"ECHO ON",
 		"Turn local echoing on",
@@ -580,7 +640,7 @@ struct cursor_type cursor_table[] =
 	{
 		"MIXED TAB BACKWARD",
 		"Tab completion on last word, search backward",
-		"[Z",
+		"[Z", // shift-tab
 		cursor_mixed_tab_backward
 	},
 	{
@@ -619,6 +679,12 @@ struct cursor_type cursor_table[] =
 		"",
 		cursor_suspend
 	},
+	{
+		"TEST",
+		"Print debugging information",
+		"",
+		cursor_test
+	},
 	{
 		"TAB BACKWARD",
 		"Tab completion from tab list, backward",
@@ -638,15 +704,13 @@ struct cursor_type cursor_table[] =
 	{
 		"", "", "[6~",   cursor_buffer_down
 	},
-	{
-		"", "", "[H",    cursor_buffer_home
-	},
-	{
-		"", "", "[F",    cursor_buffer_end
-	},
+
 	{
 		"", "", "",      cursor_buffer_lock
 	},
+	{
+		"", "", "OM",    cursor_enter
+	},
 	{
 		"", "", "[7~",   cursor_home
 	},
@@ -656,6 +720,9 @@ struct cursor_type cursor_table[] =
 	{
 		"", "", "OH",    cursor_home
 	},
+	{
+		"", "", "[H",    cursor_home
+	},
 	{
 		"", "", "OD",    cursor_left
 	},
@@ -671,6 +738,9 @@ struct cursor_type cursor_table[] =
 	{
 		"", "", "OF",    cursor_end
 	},
+	{
+		"", "", "[F",    cursor_end
+	},
 	{
 		"", "", "OC",    cursor_right
 	},
@@ -716,23 +786,39 @@ struct timer_type timer_table[] =
 	{    "Update Packet Patcher"       },
 	{    "Update Chat Server"          },
 	{    "Update Terminal"             },
+	{    "Update Time Events"          },
 	{    "Update Memory"               },
 	{    "Stall Program"               }
 };
 
 struct event_type event_table[] =
 {
+	{    "DATE",                                   "Triggers on the given date."             },
+	{    "DAY",                                    "Triggers each day or given day."         },
+	{    "HOUR",                                   "Triggers each hour or given hour."       },
 	{    "END OF PATH",                            "Triggers when walking the last room."    },
 	{    "IAC ",                                   "Triggers on telopt negotiation."         },
+	{    "MAP ENTER MAP",                          "Triggers when entering the map."         },
 	{    "MAP ENTER ROOM",                         "Triggers when entering a map room."      },
+	{    "MAP EXIT MAP",                           "Triggers when exiting the map."          },
 	{    "MAP EXIT ROOM",                          "Triggers when exiting a map room."       },
+	{    "MINUTE",                                 "Triggers each minute or given minute."   },
+	{    "MONTH",                                  "Triggers each month or given month."     },
 	{    "PROGRAM START",                          "Triggers when main session starts."      },
 	{    "PROGRAM TERMINATION",                    "Triggers when main session exists."      },
 	{    "RECEIVED INPUT",                         "Triggers when new input is received."    },
 	{    "RECEIVED LINE",                          "Triggers when a new line is received."   },
+	{    "RECEIVED OUTPUT",                        "Triggers when new output is received."   },
+	{    "SCREEN RESIZE",                          "Triggers when the screen is resized."    },
+	{    "SECOND",                                 "Trigers each second or given second."    },
 	{    "SEND OUTPUT",                            "Triggers when sending output."           },
+	{    "SESSION ACTIVATED",                      "Triggers when a session is activated."   },
 	{    "SESSION CONNECTED",                      "Triggers when a new session connects."   },
+	{    "SESSION DEACTIVATED",                    "Triggers when a session is deactivated." },
 	{    "SESSION DISCONNECTED",                   "Triggers when a session disconnects."    },
+	{    "TIME",                                   "Triggers on the given time."             },
+	{    "WEEK",                                   "Triggers each week or given week."       },
+	{    "YEAR",                                   "Triggers each year or given year."       },
 	{    "",                                       ""                                        }
 };
 
@@ -742,10 +828,12 @@ struct path_type path_table[] =
 	{    "END",               path_end               },
 	{    "INSERT",            path_ins               },
 	{    "LOAD",              path_load              },
-	{    "MAP",               path_map               },
+	{    "MAP",               path_show              },
 	{    "NEW",               path_new               },
 	{    "RUN",               path_run               },
 	{    "SAVE",              path_save              },
+	{    "SHOW",              path_show              },
+	{    "UNZIP",             path_unzip             },
 	{    "WALK",              path_walk              },
 	{    "ZIP",               path_zip               },
 	{    "",                  NULL                   }
@@ -754,8 +842,11 @@ struct path_type path_table[] =
 struct line_type line_table[] =
 {
 	{    "GAG",               line_gag               },
+	{    "IGNORE",            line_ignore            },
 	{    "LOG",               line_log               },
 	{    "LOGVERBATIM",       line_logverbatim       },
+	{    "SUBSTITUTE",        line_substitute        },
+	{    "VERBOSE",           line_verbose           },
 	{    "",                  NULL                   }
 };
 
@@ -991,7 +1082,7 @@ struct telopt_type telopt_table[] =
 	{    "198",               TEL_N,               NEG_U },
 	{    "199",               TEL_N,               NEG_U },
 	{    "200",               TEL_N,               NEG_U }, /* Used by Achaea */
-	{    "201",               TEL_N,               NEG_U },
+	{    "GMCP",              TEL_N,               NEG_U }, /* Generic Mud Communication Protocol */
 	{    "202",               TEL_N,               NEG_U },
 	{    "203",               TEL_N,               NEG_U },
 	{    "204",               TEL_N,               NEG_U },
@@ -1047,3 +1138,9 @@ struct telopt_type telopt_table[] =
 	{    "254",               TEL_N,               NEG_U },
 	{    "255",               TEL_N,               NEG_U }
 };
+
+struct term_type term_table[] =
+{
+	{    "UNKNOWN",               3 }, // ANSI + VT100
+	{    "XTERM",                59 }
+};
