@@ -31,12 +31,13 @@
 /* what we see from the remote end */
 
 char iac_do_naws[]            = { IAC, DO,    TELOPT_NAWS            };
-char iac_will_naws[]          = { IAC, WILL,  TELOPT_NAWS            };
 
 char iac_do_sga[]             = { IAC, DO,    TELOPT_SGA             };
 char iac_will_sga[]           = { IAC, WILL,  TELOPT_SGA             };
 
 char iac_sb_tspeed[]          = { IAC, SB,    TELOPT_TSPEED, ENV_SEND, IAC, SE };
+
+char iac_dont_ttype[]         = { IAC, DONT,  TELOPT_TTYPE           };
 char iac_sb_ttype[]           = { IAC, SB,    TELOPT_TTYPE,  ENV_SEND, IAC, SE  };
 
 
@@ -50,6 +51,8 @@ char iac_sb_mssp[]            = { IAC, SB,    TELOPT_MSSP            };
 
 char iac_sb_msdp[]            = { IAC, SB,    TELOPT_MSDP            };
 
+char iac_sb_gmcp[]            = { IAC, SB,    TELOPT_GMCP            };
+
 char iac_sb_new_environ[]     = { IAC, SB,    TELOPT_NEW_ENVIRON     };
 
 char iac_sb_zmp[]             = { IAC, SB,    TELOPT_ZMP             };
@@ -73,8 +76,7 @@ struct iac_type iac_table [] =
 {
 	{   3,  iac_do_sga,           &send_will_sga           },
 	{   3,  iac_will_sga,         &send_do_sga             },
-	{   3,  iac_do_naws,          &send_sb_naws            },
-	{   3,  iac_will_naws,        &send_sb_naws            },
+	{   3,  iac_do_naws,          &recv_do_naws            },
 	{   3,  iac_do_echo,          &send_echo_will          },
 	{   3,  iac_will_echo,        &send_echo_off           },
 	{   3,  iac_wont_echo,        &send_echo_on            },
@@ -82,8 +84,10 @@ struct iac_type iac_table [] =
 	{   3,  iac_will_mssp,        &send_do_mssp            },
 	{   3,  iac_sb_mssp,          &recv_sb_mssp            },
 	{   3,  iac_sb_msdp,          &recv_sb_msdp            },
+	{   3,  iac_sb_gmcp,          &recv_sb_gmcp            },
 	{   3,  iac_sb_new_environ,   &recv_sb_new_environ     },
 	{   6,  iac_sb_tspeed,        &recv_sb_tspeed          },
+	{   3,  iac_dont_ttype,       &recv_dont_ttype         },
 	{   6,  iac_sb_ttype,         &recv_sb_ttype           },
 	{   3,  iac_sb_zmp,           &recv_sb_zmp             },
 	{   5,  iac_sb_mccp1,         &init_mccp               },
@@ -116,8 +120,13 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 	unsigned char *cpdst;
 	unsigned char *cpsrc;
 
+	push_call("translate_telopts(%p,%p,%d)",ses,src,cplen);
+
 	if (cplen == 0)
 	{
+		gtd->mud_output_buf[gtd->mud_output_len] = 0;	
+
+		pop_call();
 		return 0;
 	}
 
@@ -149,15 +158,20 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 				break;
 
 			case Z_STREAM_END:
-				tintin_puts2(ses, "");
+				tintin_printf2(ses, "");
 				tintin_printf2(ses, "#COMPRESSION END, DISABLING MCCP.");
 
-				cpsrc = src + (cplen - ses->mccp->avail_in);
+				cnt  = ses->mccp->next_out - gtd->mccp_buf;
+				skip = cplen - ses->mccp->avail_in;
+
 				cplen = ses->mccp->avail_in;
+				cpsrc = src + skip;
 
 				inflateEnd(ses->mccp);
 				free(ses->mccp);
 				ses->mccp = NULL;
+
+				translate_telopts(ses, gtd->mccp_buf, cnt);
 				break;
 
 			default:
@@ -177,6 +191,12 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 		cpsrc = src;
 	}
 
+	if (HAS_BIT(ses->flags, SES_FLAG_LOGLEVEL) && ses->logfile)
+	{
+		fwrite(cpsrc, 1, cplen, ses->logfile);
+		fflush(ses->logfile);
+	}
+
  	if (ses->read_len + cplen >= ses->read_max)
 	{
 		ses->read_max = ses->read_len + cplen + 1000;
@@ -186,7 +206,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 	memcpy(ses->read_buf + ses->read_len, cpsrc, cplen);
 
 	cpsrc = ses->read_buf;
-
+	cplen = ses->read_len + cplen;
 
 	if (gtd->mud_output_len + cplen >= gtd->mud_output_max)
 	{
@@ -196,13 +216,6 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 
 	cpdst = (unsigned char *) gtd->mud_output_buf + gtd->mud_output_len;
 
-
-	if (HAS_BIT(ses->flags, SES_FLAG_LOGLEVEL) && ses->logfile)
-	{
-		fwrite(cpsrc, 1, cplen, ses->logfile);
-		fflush(ses->logfile);
-	}
-
 	while (cplen > 0)
 	{
 		if (*cpsrc == IAC)
@@ -265,7 +278,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 			{
 				if (cplen < iac_table[cnt].size && !memcmp(cpsrc, iac_table[cnt].code, cplen))
 				{
-					skip = iac_table[cnt].size;
+					skip = iac_table[cnt].size; // broken packet handling
 
 					break;
 				}
@@ -276,6 +289,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 
 					if (iac_table[cnt].func == init_mccp)
 					{
+						pop_call();
 						return translate_telopts(ses, cpsrc + skip, cplen - skip);
 					}
 					break;
@@ -303,7 +317,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 					case WILL:
 						if (cplen > 2)
 						{
-							if (!check_all_events(ses, 1, 0, "IAC WILL %s", telopt_table[cpsrc[2]].name))
+							if (!check_all_events(ses, SUB_ARG|SUB_SEC, 1, 0, "IAC WILL %s", telopt_table[cpsrc[2]].name))
 							{
 								if (!HAS_BIT(ses->telopt_flag[cpsrc[2] / 32], 1 << cpsrc[2] % 32))
 								{
@@ -325,7 +339,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 					case DO:
 						if (cplen > 2)
 						{
-							if (!check_all_events(ses, 1, 0, "IAC DO %s", telopt_table[cpsrc[2]].name))
+							if (!check_all_events(ses, SUB_ARG|SUB_SEC, 1, 0, "IAC DO %s", telopt_table[cpsrc[2]].name))
 							{
 								if (!HAS_BIT(ses->telopt_flag[cpsrc[2] / 32], 1 << cpsrc[2] % 32))
 								{
@@ -348,7 +362,9 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 					case DONT:
 						if (cplen > 2)
 						{
-							check_all_events(ses, 2, 0, "IAC %s %s", TELCMD(cpsrc[1]), telopt_table[cpsrc[2]].name);
+							check_all_events(ses, SUB_ARG|SUB_SEC, 2, 0, "IAC %s %s", TELCMD(cpsrc[1]), telopt_table[cpsrc[2]].name);
+
+							DEL_BIT(ses->telopt_flag[cpsrc[2] / 32], 1 << cpsrc[2] % 32);
 						}
 						skip = 3;
 						break;
@@ -380,6 +396,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 
 				gtd->mud_output_buf[gtd->mud_output_len] = 0;
 
+				pop_call();
 				return cplen;
 			}
 		}
@@ -411,8 +428,14 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 					cplen--;
 					continue;
 
+				case 5:
+					check_all_events(ses, SUB_ARG, 0, 1, "VT100 ENQ", gtd->term);
+					cpsrc++;
+					cplen--;
+					continue;
+
 				case '\r':
-					if (cplen < 2 || cpsrc[1] != 0)
+					if (cplen > 1 && cpsrc[1] == '\n')
 					{
 						cpsrc++;
 						cplen--;
@@ -438,6 +461,51 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 					continue;
 
 				default:
+					if (cpsrc[0] == ESCAPE)
+					{
+						if (cplen >= 2 && cpsrc[1] == 'Z')
+						{
+							check_all_events(ses, SUB_ARG, 0, 0, "VT100 DECID");
+							cpsrc += 2;
+							cplen -= 2;
+							continue;
+						}
+						if (cplen >= 3 && cpsrc[1] == '[')
+						{
+							if (cpsrc[2] == 'c')
+							{
+								check_all_events(ses, SUB_ARG, 0, 0, "VT100 DA");
+								cpsrc += 3;
+								cplen -= 3;
+								continue;
+							}
+							if (cplen >= 4)
+							{
+								if (cpsrc[2] >= '5' && cpsrc[2] <= '6' && cpsrc[3] == 'n')
+								{
+									if (cpsrc[2] == '5')
+									{
+										check_all_events(ses, SUB_ARG, 0, 0, "VT100 DSR");
+									}
+									if (cpsrc[2] == '6')
+									{
+										check_all_events(ses, SUB_ARG, 0, 2, "VT100 CPR", ntos(ses->cols), ntos(ses->rows));
+									}
+									cpsrc += 4;
+									cplen -= 4;
+									continue;
+								}
+								if (cpsrc[2] == '0' && cpsrc[3] == 'c')
+								{
+									check_all_events(ses, SUB_ARG, 0, 0, "VT100 DA");
+									cpsrc += 4;
+									cplen -= 4;
+									continue;
+								}
+							}
+						}
+					}
+
 					if (HAS_BIT(ses->telopts, TELOPT_FLAG_PROMPT))
 					{
 						DEL_BIT(ses->telopts, TELOPT_FLAG_PROMPT);
@@ -462,6 +530,7 @@ int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 
 	gtd->mud_output_buf[gtd->mud_output_len] = 0;
 
+	pop_call();
 	return 0;
 }
 
@@ -508,21 +577,46 @@ int mark_prompt(struct session *ses, int cplen, unsigned char *cpsrc)
 	TTYPE
 */
 
+int recv_dont_ttype(struct session *ses, int cplen, unsigned char *cpsrc)
+{
+	if (!check_all_events(ses, SUB_ARG|SUB_SEC, 0, 0, "IAC DONT TTYPE"))
+	{
+		DEL_BIT(ses->telopts, TELOPT_FLAG_TTYPE);
+
+		DEL_BIT(ses->telopt_flag[cpsrc[2] / 32], 1 << cpsrc[2] % 32);
+	}
+	return 3;
+}
+
 int recv_sb_ttype(struct session *ses, int cplen, unsigned char *cpsrc)
 {
-	if (!check_all_events(ses, 0, 1, "IAC SB TERMINAL TYPE", ntos(cpsrc[3])))
-	{        
-		if (HAS_BIT(ses->telopts, TELOPT_FLAG_TSPEED) && atoi(ses->port) == TELNET_PORT && getenv("TERM"))
+	if (!check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "IAC SB TTYPE", ntos(cpsrc[3])))
+	{
+		if (HAS_BIT(ses->telopts, TELOPT_FLAG_MTTS))
 		{
-			socket_printf(ses, 6 + strlen(getenv("TERM")), "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0, getenv("TERM"), IAC, SE);
+			char mtts[BUFFER_SIZE];
 
-			telopt_debug(ses, "SENT IAC SB TTYPE %s", getenv("TERM"));
+			sprintf(mtts, "MTTS %d", 1 + (HAS_BIT(ses->flags, SES_FLAG_SPLIT) ? 0 : 2) + (HAS_BIT(ses->flags, SES_FLAG_UTF8) ? 4 : 0) + (HAS_BIT(ses->flags, SES_FLAG_256COLOR) ? 8 : 0));
+
+			socket_printf(ses, 6 + strlen(mtts), "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0, mtts, IAC, SE);
+
+			telopt_debug(ses, "SENT IAC SB TTYPE %s", mtts);
+		}
+		else if (HAS_BIT(ses->telopts, TELOPT_FLAG_TTYPE))
+		{
+			socket_printf(ses, 6 + strlen(gtd->term), "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0, gtd->term, IAC, SE);
+
+			telopt_debug(ses, "SENT IAC SB TTYPE %s", gtd->term);
+
+			SET_BIT(ses->telopts, TELOPT_FLAG_MTTS);
 		}
 		else
 		{
 			socket_printf(ses, 14, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0, "TINTIN++", IAC, SE);
 
 			telopt_debug(ses, "SENT IAC SB TTYPE %s", "TINTIN++");
+
+			SET_BIT(ses->telopts, TELOPT_FLAG_TTYPE);
 		}
 	}
 	return 6;
@@ -537,7 +631,7 @@ int recv_sb_tspeed(struct session *ses, int cplen, unsigned char *cpsrc)
 {
 	SET_BIT(ses->telopts, TELOPT_FLAG_TSPEED);
 
-	if (!check_all_events(ses, 0, 1, "IAC SB TSPEED", ntos(cpsrc[3])))
+	if (!check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "IAC SB TSPEED", ntos(cpsrc[3])))
 	{
 		socket_printf(ses, 17, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TSPEED, 0, "38400,38400", IAC, SE);
 
@@ -551,32 +645,35 @@ int recv_sb_tspeed(struct session *ses, int cplen, unsigned char *cpsrc)
 	NAWS
 */
 
-int send_will_naws(struct session *ses, int cplen, unsigned char *cpsrc)
+int recv_do_naws(struct session *ses, int cplen, unsigned char *cpsrc)
 {
-	SET_BIT(ses->telopt_flag[TELOPT_NAWS / 32], 1 << TELOPT_NAWS % 32);
+	if (check_all_events(ses, SUB_ARG|SUB_SEC, 0, 0, "IAC DO NAWS"))
+	{
+		return 3;
+	}
 
-	socket_printf(ses, 3, "%c%c%c", IAC, WILL, TELOPT_NAWS);
+	SET_BIT(ses->telopts, TELOPT_FLAG_NAWS);
 
-	telopt_debug(ses, "SENT IAC WILL NAWS");
+	if (!HAS_BIT(ses->telopt_flag[TELOPT_NAWS / 32], 1 << TELOPT_NAWS % 32))
+	{
+		SET_BIT(ses->telopt_flag[TELOPT_NAWS / 32], 1 << TELOPT_NAWS % 32);
 
-	return 3;
+		socket_printf(ses, 3, "%c%c%c", IAC, WILL, TELOPT_NAWS);
+
+		telopt_debug(ses, "SENT IAC WILL NAWS");
+	}
+
+	return send_sb_naws(ses, cplen, cpsrc);
 }
 
 int send_sb_naws(struct session *ses, int cplen, unsigned char *cpsrc)
 {
 	int rows;
 
-	if (!HAS_BIT(ses->telopt_flag[TELOPT_NAWS / 32], 1 << TELOPT_NAWS % 32))
-	{
-		send_will_naws(ses, cplen, cpsrc);
-	}
-
 	rows = HAS_BIT(ses->flags, SES_FLAG_SPLIT) ? ses->bot_row - ses->top_row + 1 : ses->rows;
 
 	socket_printf(ses, 9, "%c%c%c%c%c%c%c%c%c", IAC, SB, TELOPT_NAWS, ses->cols / 256, ses->cols % 256, rows / 256, rows % 256, IAC, SE);
 
-	SET_BIT(ses->telopts, TELOPT_FLAG_NAWS);
-
 	telopt_debug(ses, "SENT IAC SB NAWS %d %d %d %d", ses->cols / 256, ses->cols % 256, ses->rows / 256, ses->rows % 256);
 
 	return 3;
@@ -679,7 +776,7 @@ int send_do_telopt(struct session *ses, int cplen, unsigned char *cpsrc)
 
 int send_do_mssp(struct session *ses, int cplen, unsigned char *cpsrc)
 {
-	if (!check_all_events(ses, 0, 0, "IAC WILL MSSP"))
+	if (!check_all_events(ses, SUB_ARG|SUB_SEC, 0, 0, "IAC WILL MSSP"))
 	{
 		if (HAS_BIT(ses->telopts, TELOPT_FLAG_DEBUG))
 		{
@@ -731,7 +828,8 @@ int recv_sb_mssp(struct session *ses, int cplen, unsigned char *src)
 
 				telopt_debug(ses, "MSSP VAR %-20s VAL %s", var, val);
 
-				check_all_events(ses, 0, 2, "IAC SB MSSP", var, val);
+				check_all_events(ses, SUB_ARG|SUB_SEC, 1, 2, "IAC SB MSSP %s", var, var, val);
+				check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "IAC SB MSSP", var, val);
 				break;
 
 			default:
@@ -742,7 +840,7 @@ int recv_sb_mssp(struct session *ses, int cplen, unsigned char *src)
 
 	telopt_debug(ses, "IAC SB MSSP IAC SE");
 
-	check_all_events(ses, 0, 0, "IAC SB MSSP IAC SE");
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 0, "IAC SB MSSP IAC SE");
 
 	return UMIN(i + 1, cplen);
 }
@@ -753,66 +851,183 @@ int recv_sb_mssp(struct session *ses, int cplen, unsigned char *src)
 
 int recv_sb_msdp(struct session *ses, int cplen, unsigned char *src)
 {
-	char var[BUFFER_SIZE], val[BUFFER_SIZE];
-	char *pto;
-	int i;
+	char var[BUFFER_SIZE], val[BUFFER_SIZE], *pto;
+	int i, nest, state[100], last;
 
-	var[0] = val[0] = i = 0;
+	var[0] = val[0] = state[0] = nest = last = 0;
 
 	if (skip_sb(ses, cplen, src) > cplen)
 	{
 		return cplen + 1;
 	}
 
-	while (i < cplen && src[i] != SE)
+	i = 3;
+	pto = var;
+
+	while (i < cplen && nest < 99)
 	{
+		if (src[i] == IAC && i + 1 < cplen && src[i+1] == SE)
+		{
+			break;
+		}
+
 		switch (src[i])
 		{
-			case MSSP_VAR:
-				i++;
-				pto = var;
-
-				while (i < cplen && src[i] >= 3 && src[i] != IAC)
-				{
-					*pto++ = src[i++];
-				}
-				*pto = 0;
+			case MSDP_TABLE_OPEN:
+				nest++;
+				state[nest] = 0;
+				last = MSDP_TABLE_OPEN;
 				break;
 
-			case MSSP_VAL:
-				i++;
-				pto = val;
-
-				while (i < cplen && src[i] >= 3 && src[i] != IAC)
+			case MSDP_TABLE_CLOSE:
+				if (nest)
 				{
-					*pto++ = src[i++];
+					if (last == MSDP_VAL || last == MSDP_VAR)
+					{
+						*pto++ = '}';
+					}
+					nest--;
 				}
-				*pto = 0;
+				if (nest)
+				{
+					*pto++ = '}';
+				}
+				last = MSDP_TABLE_CLOSE;
+				break;
 
-				telopt_debug(ses, "IAC SB MSDP (VAR %-20s VAL %s)", var, val);
+			case MSDP_ARRAY_OPEN:
+				nest++;
+				state[nest] = 1;
+				last = MSDP_ARRAY_OPEN;
+				break;
 
-				check_all_events(ses, 0, 2, "IAC SB MSDP", var, val);
+			case MSDP_ARRAY_CLOSE:
+				if (nest)
+				{
+					if (last == MSDP_VAL)
+					{
+						*pto++ = '}';
+					}
+					nest--;
+				}
+				if (nest)
+				{
+					*pto++ = '}';
+				}
+				last = MSDP_ARRAY_CLOSE;
+				break;
+
+			case MSDP_VAR:
+				if (nest)
+				{
+					if (last == MSDP_VAL)
+					{
+						*pto++ = '}';
+					}
+					*pto++ = '{';
+				}
+				else
+				{
+					*pto = 0;
+
+					if (last)
+					{
+						telopt_debug(ses, "IAC SB MSDP VAR %-20s VAL %s", var, val);
+						check_all_events(ses, SUB_ARG, 1, 2, "IAC SB MSDP %s", var, var, val);
+						check_all_events(ses, SUB_ARG, 0, 2, "IAC SB MSDP", var, val);
+					}
+					pto = var;
+				}
+				last = MSDP_VAR;
+				break;
+
+			case MSDP_VAL:
+				if (nest)
+				{
+					if (last == MSDP_VAR || last == MSDP_VAL)
+					{
+						*pto++ = '}';
+					}
+					if (state[nest])
+					{
+						pto += sprintf(pto, "{%d}", state[nest]++);
+					}
+					*pto++ = '{';
+				}
+				else
+				{
+					*pto = 0;
+
+					if (last != MSDP_VAR)
+					{
+						telopt_debug(ses, "IAC SB MSDP VAR %-20s VAL %s", var, val);
+						check_all_events(ses, SUB_ARG, 1, 2, "IAC SB MSDP %s", var, var, val);
+						check_all_events(ses, SUB_ARG, 0, 2, "IAC SB MSDP", var, val);
+					}
+					pto = val;
+				}
+				last = MSDP_VAL;
+				break;
+
+			case '\\':
+				*pto++ = '\\';
+				*pto++ = '\\';
+				break;
+
+			case '{':
+				*pto++ = '\\';
+				*pto++ = 'x';
+				*pto++ = '7';
+				*pto++ = 'B';
+				break;
+
+			case '}':
+				*pto++ = '\\';
+				*pto++ = 'x';
+				*pto++ = '7';
+				*pto++ = 'D';
+				break;
+
+			case COMMAND_SEPARATOR:
+				*pto++ = '\\';
+				*pto++ = COMMAND_SEPARATOR;
 				break;
 
 			default:
-				i++;
+				*pto++ = src[i];
 				break;
 		}
+		i++;
+	}
+
+	if (i < cplen && src[i] == IAC && nest < 99)
+	{
+		*pto = 0;
+
+		if (last)
+		{
+			telopt_debug(ses, "IAC SB MSDP VAR %-20s VAL %s", var, val);
+			check_all_events(ses, SUB_ARG, 1, 2, "IAC SB MSDP %s", var, var, val);
+			check_all_events(ses, SUB_ARG, 0, 2, "IAC SB MSDP", var, val);
+		}
+		i++;
 	}
 
 	telopt_debug(ses, "IAC SB MSDP IAC SE");
 
-	check_all_events(ses, 0, 0, "IAC SB MSDP IAC SE");
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 0, "IAC SB MSDP IAC SE");
+
 	return UMIN(i + 1, cplen);
 }
 
+
 /*
 	NEW-ENVIRON
 */
 
 int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src)
 {
-	char var[BUFFER_SIZE], val[BUFFER_SIZE];
+	char buf[BUFFER_SIZE], var[BUFFER_SIZE], val[BUFFER_SIZE];
 	char *pto;
 	int i, x;
 
@@ -836,7 +1051,7 @@ int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src)
 			case ENV_USR:
 				x = src[i];
 				i++;
-				pto = var;
+				pto = buf;
 
 				while (i < cplen && src[i] >= 4 && src[i] != IAC)
 				{
@@ -844,17 +1059,19 @@ int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src)
 				}
 				*pto = 0;
 
+				substitute(ses, buf, var, SUB_SEC);
+
 				if (src[3] == ENV_SEND)
 				{
 					telopt_debug(ses, "IAC SB NEW-ENVIRON SEND %s", x ? "VAR" : "USR");
 
-					check_all_events(ses, 1, 2, "IAC SB NEW-ENVIRON SEND %s", x ? "VAR" : "USR", var, "");
+					check_all_events(ses, SUB_ARG|SUB_SEC, 1, 2, "IAC SB NEW-ENVIRON SEND %s", x ? "VAR" : "USR", var, "");
 				}
 				break;
 
 			case ENV_VAL:
 				i++;
-				pto = val;
+				pto = buf;
 
 				while (i < cplen && src[i] >= 3 && src[i] != IAC)
 				{
@@ -862,9 +1079,11 @@ int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src)
 				}
 				*pto = 0;
 
+				substitute(ses, buf, val, SUB_SEC);
+
 				telopt_debug(ses, "IAC SB NEW-ENVIRON %s %s", src[3] ? "IS" : "INFO", x ? "VAR" : "USR");
 
-				check_all_events(ses, 2, 2, "IAC SB NEW-ENVIRON %s %s", src[i] ? "IS" : "INFO", x ? "VAR" : "USR", var, val);
+				check_all_events(ses, SUB_ARG|SUB_SEC, 2, 2, "IAC SB NEW-ENVIRON %s %s", src[i] ? "IS" : "INFO", x ? "VAR" : "USR", var, val);
 				break;
 
 			case IAC:
@@ -872,7 +1091,7 @@ int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src)
 				{
 					telopt_debug(ses, "IAC SB NEW-ENVIRON %s", src[3] == 0 ? "IS" : src[3] == 1 ? "SEND" : "INFO");
 
-					check_all_events(ses, 1, 2, "IAC SB NEW-ENVIRON %s", src[3] == 0 ? "IS" : src[3] == 1 ? "SEND" : "INFO", var, val);
+					check_all_events(ses, SUB_ARG|SUB_SEC, 1, 2, "IAC SB NEW-ENVIRON %s", src[3] == 0 ? "IS" : src[3] == 1 ? "SEND" : "INFO", var, val);
 				}
 				i++;
 				break;
@@ -884,14 +1103,14 @@ int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src)
 
 	telopt_debug(ses, "IAC SB NEW-ENVIRON IAC SE");
 
-	check_all_events(ses, 0, 0, "IAC SB NEW-ENVIRON IAC SE");
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 0, "IAC SB NEW-ENVIRON IAC SE");
 
 	return i + 1;
 }
 
 int recv_sb_zmp(struct session *ses, int cplen, unsigned char *src)
 {
-	char var[BUFFER_SIZE], val[BUFFER_SIZE];
+	char buf[BUFFER_SIZE], var[BUFFER_SIZE], val[BUFFER_SIZE];
 	char *pto;
 	int i, x;
 
@@ -913,7 +1132,7 @@ int recv_sb_zmp(struct session *ses, int cplen, unsigned char *src)
 				break;
 
 			default:
-				pto = x ? val : var;
+				pto = buf;
 
 				while (i < cplen && src[i])
 				{
@@ -921,11 +1140,13 @@ int recv_sb_zmp(struct session *ses, int cplen, unsigned char *src)
 				}
 				*pto = src[i++];
 
+				substitute(ses, buf, x ? val : var, SUB_SEC);
+
 				if (x++)
 				{
 					telopt_debug(ses, "IAC SB ZMP %s", var);
 
-					check_all_events(ses, 1, 1, "IAC SB ZMP %s", var, val);
+					check_all_events(ses, SUB_ARG|SUB_SEC, 1, 1, "IAC SB ZMP %s", var, val);
 				}
 				break;
 		}
@@ -933,11 +1154,269 @@ int recv_sb_zmp(struct session *ses, int cplen, unsigned char *src)
 
 	telopt_debug(ses, "IAC SB ZMP %s IAC SE", var);
 
-	check_all_events(ses, 1, 0, "IAC SB ZMP %s IAC SE", var);
+	check_all_events(ses, SUB_ARG|SUB_SEC, 1, 0, "IAC SB ZMP %s IAC SE", var);
 
 	return UMIN(i + 1, cplen);
 }
 
+int recv_sb_gmcp(struct session *ses, int cplen, unsigned char *src)
+{
+	char mod[BUFFER_SIZE], val[BUFFER_SIZE], json[BUFFER_SIZE], *pto;
+	int i, state[100], nest, type;
+
+	push_call("recv_sb_gmcp(%p,%d,%p)",ses,cplen,src);
+
+	if (skip_sb(ses, cplen, src) > cplen)
+	{
+		pop_call();
+		return cplen + 1;
+	}
+
+	mod[0] = val[0] = state[0] = nest = type = 0;
+
+	i = 3;
+
+	pto = mod;
+
+	// space out
+
+	while (i < cplen && src[i] == ' ')
+	{
+		i++;
+	}
+
+	// grab module
+
+	while (i < cplen && src[i] != IAC)
+	{
+		if (src[i] == ' ' || src[i] == '{' || src[i] == '[')
+		{
+			break;
+		}
+		*pto++ = src[i++];
+	}
+
+	*pto = 0;
+
+	// parse JSON content
+
+	pto = val;
+
+	while (i < cplen && src[i] != IAC && nest < 99)
+	{
+		switch (src[i])
+		{
+			case ' ':
+				i++;
+				break;
+
+			case '{':
+				if (nest != 0)
+				{
+					*pto++ = '{';
+				}
+				i++;
+				state[++nest] = 0;
+				break;
+
+			case '}':
+				nest--;
+				i++;
+				if (nest != 0)
+				{
+					*pto++ = '}';
+				}
+				break;
+
+			case '[':
+				if (nest != 0)
+				{
+					*pto++ = '{';
+				}
+				i++;
+				state[++nest] = 1;
+				pto += sprintf(pto, "{%d}", state[nest]);
+				break;
+
+			case ']':
+				nest--;
+				i++;
+				if (nest != 0)
+				{
+					*pto++ = '}';
+				}
+				break;
+
+			case ':':
+				i++;
+				break;
+
+			case ',':
+				i++;
+				if (state[nest])
+				{
+					pto += sprintf(pto, "{%d}", ++state[nest]);
+				}
+				break;
+
+			case '"':
+				i++;
+				if (nest)
+				{
+					*pto++ = '{';
+				}
+				type = 1;
+
+				while (i < cplen && src[i] != IAC && type == 1)
+				{
+					switch (src[i])
+					{
+						case '\\':
+							i++;
+
+							if (i < cplen && src[i] == '"')
+							{
+								*pto++ = src[i++];
+							}
+							else
+							{
+								*pto++ = '\\';
+							}
+							break;
+
+						case '"':
+							i++;
+							type = 0;
+							break;
+
+						case '{':
+							i++;
+							*pto++ = '\\';
+							*pto++ = 'x';
+							*pto++ = '7';
+							*pto++ = 'B';
+							break;
+
+						case '}':
+							i++;
+							*pto++ = '\\';
+							*pto++ = 'x';
+							*pto++ = '7';
+							*pto++ = 'D';
+							break;
+
+						case COMMAND_SEPARATOR:
+							i++;
+							*pto++ = '\\';
+							*pto++ = COMMAND_SEPARATOR;
+							break;
+
+						default:
+							*pto++ = src[i++];
+							break;
+					}
+				}
+
+				if (nest)
+				{
+					*pto++ = '}';
+				}
+				break;
+
+			default:
+				if (nest)
+				{
+					*pto++ = '{';
+				}
+
+				type = 1;
+
+				while (i < cplen && src[i] != IAC && type == 1)
+				{
+					switch (src[i])
+					{
+						case '}':
+						case ']':
+						case ',':
+						case ':':
+							type = 0;
+							break;
+
+						case ' ':
+							i++;
+							break;
+
+						default:
+							*pto++ = src[i++];
+							break;
+					}
+				}
+
+				if (nest)
+				{
+					*pto++ = '}';
+				}
+				break;
+		}
+	}
+	*pto = 0;
+
+	// Raw json data for debugging purposes.
+
+	pto = json;
+	i = 3;
+
+	while (i < cplen && src[i] != IAC)
+	{
+		switch (src[i])
+		{
+			case '\\':
+				i++;
+				*pto++ = '\\';
+				*pto++ = '\\';
+				break;
+
+			case '{':
+				i++;
+				*pto++ = '\\';
+				*pto++ = 'x';
+				*pto++ = '7';
+				*pto++ = 'B';
+				break;
+
+			case '}':
+				i++;
+				*pto++ = '\\';
+				*pto++ = 'x';
+				*pto++ = '7';
+				*pto++ = 'D';
+				break;
+
+			case COMMAND_SEPARATOR:
+				i++;
+				*pto++ = '\\';
+				*pto++ = COMMAND_SEPARATOR;
+				break;
+
+			default:
+				*pto++ = src[i++];
+				break;
+		}
+	}
+	*pto = 0;
+
+	while (i < cplen && src[i] != SE)
+	{
+		i++;
+	}
+
+	telopt_debug(ses, "IAC SB GMCP %s IAC SE", mod);
+
+	check_all_events(ses, SUB_ARG, 1, 2, "IAC SB GMCP %s IAC SE", mod, val, json);
+
+	pop_call();
+	return UMIN(i + 1, cplen);
+}
 
 /*
 	MCCP
@@ -1010,12 +1489,6 @@ void zlib_free(void *opaque, void *address)
 	free(address);
 }
 
-void init_telnet_session(struct session *ses)
-{
-//	send_do_sga(ses, 0, NULL);
-//	send_will_naws(ses, 0, NULL);
-}
-
 
 /*
 	Returns the length of a telnet subnegotiation
@@ -1033,6 +1506,8 @@ int skip_sb(struct session *ses, int cplen, unsigned char *cpsrc)
 		}
 	}
 
+	telopt_debug(ses, "SKIP SB (%d)", cplen);
+
 	return cplen + 1;
 }
 
@@ -1068,7 +1543,7 @@ int recv_sb(struct session *ses, int cplen, unsigned char *cpsrc)
 	*pt1 = 0;
 	*pt2 = 0;
 
-	check_all_events(ses, 1, 2, "IAC SB %s", telopt_table[cpsrc[2]].name, var1, var2);
+	check_all_events(ses, SUB_ARG|SUB_SEC, 1, 2, "IAC SB %s", telopt_table[cpsrc[2]].name, var1, var2);
 
 	return i + 2;
 }
