@@ -93,16 +93,14 @@ int word_wrap(struct session *ses, char *textin, char *textout, int display)
 	{
 		if (skip_vt102_codes(pti))
 		{
-			if (display == FALSE || interpret_vt102_codes(ses, pti, TRUE))
+			if (display)
 			{
-				for (skip = skip_vt102_codes(pti) ; skip > 0 ; skip--)
-				{
-					*pto++ = *pti++;
-				}
+				interpret_vt102_codes(ses, pti, TRUE);
 			}
-			else
+
+			for (skip = skip_vt102_codes(pti) ; skip > 0 ; skip--)
 			{
-				pti += skip_vt102_codes(pti);
+				*pto++ = *pti++;
 			}
 			continue;
 		}
@@ -156,9 +154,161 @@ int word_wrap(struct session *ses, char *textin, char *textout, int display)
 		}
 		else
 		{
-			ses->cur_col++;
+			if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+			{
+				*pto++ = *pti++;
+				*pto++ = *pti++;
+			}
+			else if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (*pti & 192) == 192)
+			{
+				*pto++ = *pti++;
 
-			*pto++ = *pti++;
+				while ((*pti & 192) == 128)
+				{
+					*pto++ = *pti++;
+				}
+			}
+			else
+			{
+				*pto++ = *pti++;
+			}
+			ses->cur_col++;
+		}
+	}
+	*pto = 0;
+
+	pop_call();
+	return (cnt + 1);
+}
+
+// store whatever falls inbetween skip and keep.
+
+int word_wrap_split(struct session *ses, char *textin, char *textout, int skip, int keep)
+{
+	char *pti, *pto, *lis, *los, *chi, *cho, *ptb;
+	int i = 0, cnt = 0;
+
+	push_call("word_wrap_split(%s,%p,%p,%d,%d)",ses->name, textin,textout, skip, keep);
+
+	pti = chi = lis = textin;
+	pto = cho = los = textout;
+
+	ses->cur_col = 1;
+
+	while (*pti != 0)
+	{
+		if (skip_vt102_codes(pti))
+		{
+			if (cnt >= skip && cnt < keep)
+			{
+				for (i = skip_vt102_codes(pti) ; i > 0 ; i--)
+				{
+					*pto++ = *pti++;
+				}
+			}
+			else
+			{
+				pti += skip_vt102_codes(pti);
+			}
+			continue;
+		}
+
+		if (*pti == '\n')
+		{
+			if (cnt >= skip && cnt < keep)
+			{
+				*pto++ = *pti++;
+			}
+			else
+			{
+				pti++;
+			}
+			cnt = cnt + 1;
+			los = pto;
+			lis = pti;
+
+			ses->cur_col = 1;
+
+			continue;
+		}
+
+		if (*pti == ' ')
+		{
+			los = pto;
+			lis = pti;
+		}
+
+		if (ses->cur_col > ses->cols)
+		{
+			cnt++;
+			ses->cur_col = 1;
+
+			if (HAS_BIT(ses->flags, SES_FLAG_WORDWRAP))
+			{
+				if (pto - los > 15 || !SCROLL(ses))
+				{
+					if (cnt >= skip && cnt < keep)
+					{
+						*pto++ = '\n';
+					}
+					los = pto;
+					lis = pti;
+				}
+				else if (lis != chi) // infinite VT loop detection
+				{
+					if (cnt >= skip && cnt < keep)
+					{
+						pto = los;
+						*pto++ = '\n';
+					}
+					pti = chi = lis;
+					pti++;
+				}
+				else if (los != cho)
+				{
+					if (cnt >= skip && cnt < keep)
+					{
+						pto = cho = los;
+						pto++;
+					}
+					else
+					{
+						cho = los;
+					}
+					pti = chi = lis;
+					pti++;
+				}
+			}
+		}
+		else
+		{
+			ptb = pto;
+
+			if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+			{
+				*pto++ = *pti++;
+				*pto++ = *pti++;
+			}
+			else if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (*pti & 192) == 192)
+			{
+				*pto++ = *pti++;
+
+				while ((*pti & 192) == 128)
+				{
+					*pto++ = *pti++;
+				}
+			}
+			else
+			{
+				*pto++ = *pti++;
+			}
+
+			if (cnt < skip || cnt >= keep)
+			{
+				pto = ptb;
+			}
+
+			ses->cur_col++;
 		}
 	}
 	*pto = 0;
