@@ -32,21 +32,15 @@
 #include "tintin.h"
 
 
-int match(struct session *ses, char *str, char *exp)
+int match(struct session *ses, char *str, char *exp, int flags)
 {
 	char expbuf[BUFFER_SIZE];
 
-	if (ses)
-	{
-		sprintf(expbuf, "^%s$", exp);
+	sprintf(expbuf, "^%s$", exp);
 
-		substitute(ses, expbuf, expbuf, SUB_VAR|SUB_FUN);
-	}
-	else
-	{
-		sprintf(expbuf, "^%s$", exp);
-	}
-	return tintin_regexp(NULL, str, expbuf, 0, 0);
+	substitute(ses, expbuf, expbuf, flags);
+
+	return tintin_regexp(ses, NULL, str, expbuf, 0, 0);
 }
 
 int find(struct session *ses, char *str, char *exp, int sub)
@@ -58,11 +52,11 @@ int find(struct session *ses, char *str, char *exp, int sub)
 		substitute(ses, str, strbuf, SUB_VAR|SUB_FUN);
 		substitute(ses, exp, expbuf, SUB_VAR|SUB_FUN);
 
-		return tintin_regexp(NULL, strbuf, expbuf, 0, sub);
+		return tintin_regexp(ses, NULL, strbuf, expbuf, 0, sub);
 	}
 	else
 	{
-		return tintin_regexp(NULL, str, exp, 0, sub);
+		return tintin_regexp(ses, NULL, str, exp, 0, sub);
 	}
 }
 
@@ -70,10 +64,10 @@ DO_COMMAND(do_regexp)
 {
 	char left[BUFFER_SIZE], right[BUFFER_SIZE], is_true[BUFFER_SIZE], is_false[BUFFER_SIZE];
 
-	arg = get_arg_in_braces(arg, left,  0);
-	arg = get_arg_in_braces(arg, right, 0);
-	arg = get_arg_in_braces(arg, is_true,  1);
-	arg = get_arg_in_braces(arg, is_false, 1);
+	arg = get_arg_in_braces(ses, arg, left,  0);
+	arg = get_arg_in_braces(ses, arg, right, 0);
+	arg = get_arg_in_braces(ses, arg, is_true,  1);
+	arg = get_arg_in_braces(ses, arg, is_false, 1);
 
 	if (*is_true == 0)
 	{
@@ -84,7 +78,7 @@ DO_COMMAND(do_regexp)
 		substitute(ses, left,  left,  SUB_VAR|SUB_FUN);
 		substitute(ses, right, right, SUB_VAR|SUB_FUN);
 
-		if (tintin_regexp(NULL, left, right, 0, SUB_CMD))
+		if (tintin_regexp(ses, NULL, left, right, 0, SUB_CMD))
 		{
 			substitute(ses, is_true, is_true, SUB_CMD);
 
@@ -202,18 +196,12 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 
 	while (TRUE)
 	{
-
-#ifdef BIG5
-		if (*pti & 0x80)
+		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
 		{
 			*pto++ = *pti++;
-			if (*pti)
-			{
-				*pto++ = *pti++;
-			}
+			*pto++ = *pti++;
 			continue;
 		}
-#endif
 
 		switch (*pti)
 		{
@@ -252,7 +240,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 				}
 
 			case '$':
-				if (HAS_BIT(flags, SUB_VAR) && (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '$'))
+				if (HAS_BIT(flags, SUB_VAR) && (pti[1] == DEFAULT_OPEN || isalpha((int) pti[1]) || pti[1] == '$'))
 				{
 					int def = FALSE;
 
@@ -263,7 +251,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 							*pto++ = *pti++;
 						}
 
-						if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]))
+						if (pti[1] == DEFAULT_OPEN || isalnum((int) pti[1]))
 						{
 							pti++;
 						}
@@ -280,7 +268,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 					{
 						def = TRUE;
 
-						pti = get_arg_in_braces(pti, buf, TRUE);
+						pti = get_arg_in_braces(ses, pti, buf, TRUE);
 
 						substitute(ses, buf, temp, flags_neol);
 					}
@@ -288,18 +276,20 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 					{
 						ptt = temp;
 
-						while (isalnum(*pti) || *pti == '_')
+						while (isalnum((int) *pti) || *pti == '_')
 						{
 							*ptt++ = *pti++;
 						}
 						*ptt = 0;
 					}
 
-					pti = get_arg_at_brackets(pti, temp + strlen(temp));
+					pti = get_arg_at_brackets(ses, pti, temp + strlen(temp));
 
-					get_nest_node(ses->list[LIST_VARIABLE], temp, buf, def);
+					substitute(ses, temp, buf, flags_neol);
 
-					substitute(ses, buf, pto, flags_neol - SUB_VAR);
+					get_nest_node(ses->list[LIST_VARIABLE], buf, temp, def);
+
+					substitute(ses, temp, pto, flags_neol - SUB_VAR);
 
 					pto += strlen(pto);
 				}
@@ -312,7 +302,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 			case '<':
 				if (HAS_BIT(flags, SUB_COL))
 				{
-					if (isdigit(pti[1]) && isdigit(pti[2]) && isdigit(pti[3]) && pti[4] == '>')
+					if (isdigit((int) pti[1]) && isdigit((int) pti[2]) && isdigit((int) pti[3]) && pti[4] == '>')
 					{
 						if (pti[1] != '8' || pti[2] != '8' || pti[3] != '8')
 						{
@@ -389,7 +379,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 						*pto++ = 'm';
 						pti += 5;
 					}
-					else if (pti[1] == 'g' && isdigit(pti[2]) && isdigit(pti[3]) && pti[4] == '>')
+					else if (pti[1] == 'g' && isdigit((int) pti[2]) && isdigit((int) pti[3]) && pti[4] == '>')
 					{
 						*pto++ = ESCAPE;
 						*pto++ = '[';
@@ -405,7 +395,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 						*pto++ = 'm';
 						pti += 5;
 					}
-					else if (pti[1] == 'G' && isdigit(pti[2]) && isdigit(pti[3]) && pti[4] == '>')
+					else if (pti[1] == 'G' && isdigit((int) pti[2]) && isdigit((int) pti[3]) && pti[4] == '>')
 					{
 						*pto++ = ESCAPE;
 						*pto++ = '[';
@@ -443,7 +433,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 						continue;
 					}
 
-					for (ptt = temp, i = 1 ; isalnum(pti[i]) || pti[i] == '_' ; i++)
+					for (ptt = temp, i = 1 ; isalnum((int) pti[i]) || pti[i] == '_' ; i++)
 					{
 						*ptt++ = pti[i];
 					}
@@ -464,7 +454,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 						continue;
 					}
 
-					pti = get_arg_in_braces(&pti[i], temp, FALSE);
+					pti = get_arg_in_braces(ses, &pti[i], temp, FALSE);
 
 					substitute(ses, temp, buf, flags_neol);
 
@@ -476,7 +466,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 
 					for (i = 1 ; i < 100 ; i++)
 					{
-						pte = get_arg_in_braces(pte, temp, FALSE);
+						pte = get_arg_in_braces(ses, pte, temp, TRUE);
 
 						RESTRING(gtd->vars[i], temp);
 
@@ -489,6 +479,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 						{
 							pte++;
 						}
+
 					}
 
 					substitute(ses, node->right, buf, SUB_ARG);
@@ -506,7 +497,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 				break;
 
 			case '%':
-				if (HAS_BIT(flags, SUB_ARG) && (isdigit(pti[1]) || pti[1] == '%'))
+				if (HAS_BIT(flags, SUB_ARG) && (isdigit((int) pti[1]) || pti[1] == '%'))
 				{
 					if (pti[1] == '%')
 					{
@@ -518,23 +509,19 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 					}
 					else
 					{
-						i = isdigit(pti[2]) ? (pti[1] - '0') * 10 + pti[2] - '0' : pti[1] - '0';
+						i = isdigit((int) pti[2]) ? (pti[1] - '0') * 10 + pti[2] - '0' : pti[1] - '0';
 
 						ptt = gtd->vars[i];
 
 						while (*ptt)
 						{
-#ifdef BIG5
-							if (*ptt & 0x80)
+							if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
 							{
-								*pto++ = *ptt++;
-								if (*ptt)
-								{
-									*pto++ = *ptt++;
-								}
+								*pto++ = *pti++;
+								*pto++ = *pti++;
 								continue;
 							}
-#endif
+
 							if (HAS_BIT(flags, SUB_SEC))
 							{
 								switch (*ptt)
@@ -574,7 +561,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 								*pto++ = *ptt++;
 							}
 						}
-						pti += isdigit(pti[2]) ? 3 : 2;
+						pti += isdigit((int) pti[2]) ? 3 : 2;
 					}
 				}
 				else
@@ -584,7 +571,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 				break;
 
 			case '&':
-				if (HAS_BIT(flags, SUB_CMD) && (isdigit(pti[1]) || pti[1] == '&'))
+				if (HAS_BIT(flags, SUB_CMD) && (isdigit((int) pti[1]) || pti[1] == '&'))
 				{
 					if (pti[1] == '&')
 					{
@@ -592,7 +579,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 						{
 							*pto++ = *pti++;
 						}
-						if (isdigit(pti[1]))
+						if (isdigit((int) pti[1]))
 						{
 							pti++;
 						}
@@ -603,16 +590,16 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 					}
 					else
 					{
-						i = isdigit(pti[2]) ? (pti[1] - '0') * 10 + pti[2] - '0' : pti[1] - '0';
+						i = isdigit((int) pti[2]) ? (pti[1] - '0') * 10 + pti[2] - '0' : pti[1] - '0';
 
 						for (cnt = 0 ; gtd->cmds[i][cnt] ; cnt++)
 						{
 							*pto++ = gtd->cmds[i][cnt];
 						}
-						pti += isdigit(pti[2]) ? 3 : 2;
+						pti += isdigit((int) pti[2]) ? 3 : 2;
 					}
 				}
-				else if (HAS_BIT(flags, SUB_VAR) && (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '&'))
+				else if (HAS_BIT(flags, SUB_VAR) && (pti[1] == DEFAULT_OPEN || isalpha((int) pti[1]) || pti[1] == '&'))
 				{
 					int def = 0;
 
@@ -623,7 +610,7 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 							*pto++ = *pti++;
 						}
 
-						if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]))
+						if (pti[1] == DEFAULT_OPEN || isalnum((int) pti[1]))
 						{
 							pti++;
 						}
@@ -641,24 +628,26 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 					{
 						def = TRUE;
 
-						pti = get_arg_in_braces(pti, buf, TRUE);
+						pti = get_arg_in_braces(ses, pti, buf, TRUE);
 
 						substitute(ses, buf, temp, flags_neol);
 					}
 					else
 					{
-						for (ptt = temp ; isalnum(*pti) || *pti == '_' ; i++)
+						for (ptt = temp ; isalnum((int) *pti) || *pti == '_' ; i++)
 						{
 							*ptt++ = *pti++;
 						}
 						*ptt = 0;
 					}
 
-					pti = get_arg_at_brackets(pti, temp + strlen(temp));
+					pti = get_arg_at_brackets(ses, pti, temp + strlen(temp));
 
-					get_nest_index(ses->list[LIST_VARIABLE], temp, buf, def);
+					substitute(ses, temp, buf, flags_neol);
 
-					substitute(ses, buf, pto, flags_neol - SUB_VAR);
+					get_nest_index(ses->list[LIST_VARIABLE], buf, temp, def);
+
+					substitute(ses, temp, pto, flags_neol - SUB_VAR);
 
 					pto += strlen(pto);
 				}
@@ -733,7 +722,44 @@ int substitute(struct session *ses, char *string, char *result, int flags)
 				break;
 
 			default:
-				*pto++ = *pti++;
+				if (HAS_BIT(flags, SUB_SEC) && !HAS_BIT(flags, SUB_ARG))
+				{
+					switch (*pti)
+					{
+						case '\\':
+							*pto++ = '\\';
+							*pto++ = '\\';
+							break;
+
+						case '{':
+							*pto++ = '\\';
+							*pto++ = 'x';
+							*pto++ = '7';
+							*pto++ = 'B';
+							break;
+
+						case '}':
+							*pto++ = '\\';
+							*pto++ = 'x';
+							*pto++ = '7';
+							*pto++ = 'D';
+							break;
+
+						case COMMAND_SEPARATOR:
+							*pto++ = '\\';
+							*pto++ = COMMAND_SEPARATOR;
+							break;
+
+						default:
+							*pto++ = *pti;
+							break;
+					}
+					pti++;
+				}
+				else
+				{
+					*pto++ = *pti++;
+				}
 				break;
 		}
 	}
@@ -772,14 +798,14 @@ int check_one_regexp(struct session *ses, struct listnode *node, char *line, cha
 		str = line;
 	}	
 
-	return tintin_regexp(node->regex, str, exp, option, SUB_ARG);
+	return tintin_regexp(ses, node->regex, str, exp, option, SUB_ARG);
 }
 
 /*
 	Keep synched with tintin_regexp and tintin_regexp_compile
 */
 
-int tintin_regexp_check(char *exp)
+int tintin_regexp_check(struct session *ses, char *exp)
 {
 	if (*exp == '^')
 	{
@@ -788,13 +814,13 @@ int tintin_regexp_check(char *exp)
 
 	while (*exp)
 	{
-#ifdef BIG5
-		if (*exp & 0x80)
+
+		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *exp & 128 && exp[1] != 0)
 		{
 			exp += 2;
 			continue;
 		}
-#endif		
+
 		switch (exp[0])
 		{
 			case '\\':
@@ -843,7 +869,7 @@ int tintin_regexp_check(char *exp)
 	return FALSE;
 }
 				
-int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
+int tintin_regexp(struct session *ses, pcre *nodepcre, char *str, char *exp, int option, int flag)
 {
 	char out[BUFFER_SIZE], *pti, *pto;
 	int arg = 1, var = 1, fix = 0;
@@ -858,14 +884,13 @@ int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
 
 	while (*pti)
 	{
-#ifdef BIG5
-		if (*pti & 0x80)
+		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
 			continue;
 		}
-#endif
+
 		switch (pti[0])
 		{
 			case '\\':
@@ -876,7 +901,7 @@ int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
 			case '{':
 				gtd->args[up(var)] = up(arg);
 				*pto++ = '(';
-				pti = get_arg_in_braces(pti, pto, TRUE);
+				pti = get_arg_in_braces(ses, pti, pto, TRUE);
 				pto += strlen(pto);
 				*pto++ = ')';
 				break;
@@ -896,7 +921,7 @@ int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
 				break;
 
 			case '$':
-				if (pti[1] != DEFAULT_OPEN && !isalnum(pti[1]))
+				if (pti[1] != DEFAULT_OPEN && !isalnum((int) pti[1]))
 				{
 					int i = 0;
 
@@ -927,9 +952,9 @@ int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
 					case '8':
 					case '9':
 						fix = SUB_FIX;
-						arg = isdigit(pti[2]) ? (pti[1] - '0') * 10 + (pti[2] - '0') : pti[1] - '0';
+						arg = isdigit((int) pti[2]) ? (pti[1] - '0') * 10 + (pti[2] - '0') : pti[1] - '0';
 						gtd->args[up(var)] = up(arg);
-						pti += isdigit(pti[2]) ? 3 : 2;
+						pti += isdigit((int) pti[2]) ? 3 : 2;
 						strcpy(pto, *pti == 0 ? "(.*)" : "(.*?)");
 						pto += strlen(pto);
 						break;
@@ -991,7 +1016,7 @@ int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
 					case '?':
 						gtd->args[up(var)] = up(arg);
 						pti += 2;
-						strcpy(pto, *pti == 0 ? "(.?)" : "(.??)");
+						strcpy(pto, *pti == 0 ? "(.?)" : "(.?" "?)");
 						pto += strlen(pto);
 						break;
 
@@ -1037,7 +1062,7 @@ int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
 	return regexp_compare(nodepcre, str, out, option, flag + fix);
 }
 
-pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
+pcre *tintin_regexp_compile(struct session *ses, struct listnode *node, char *exp, int option)
 {
 	char out[BUFFER_SIZE], *pti, *pto;
 
@@ -1057,15 +1082,12 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 
 	while (*pti)
 	{
-
-#ifdef BIG5
-		if (*pti & 0x80)
+		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
 			continue;
 		}
-#endif
 		switch (pti[0])
 		{
 			case '\\':
@@ -1075,7 +1097,7 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 
 			case '{':
 				*pto++ = '(';
-				pti = get_arg_in_braces(pti, pto, TRUE);
+				pti = get_arg_in_braces(ses, pti, pto, TRUE);
 				while (*pto)
 				{
 					if (pto[0] == '$' || pto[0] == '@')
@@ -1091,7 +1113,7 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 				break;
 
 			case '&':
-				if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '&')
+				if (pti[1] == DEFAULT_OPEN || isalnum((int) pti[1]) || pti[1] == '&')
 				{
 					return NULL;
 				}
@@ -1099,7 +1121,7 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 				break;
 
 			case '@':
-				if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '@')
+				if (pti[1] == DEFAULT_OPEN || isalnum((int) pti[1]) || pti[1] == '@')
 				{
 					return NULL;
 				}
@@ -1107,7 +1129,7 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 				break;
 
 			case '$':
-				if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]))
+				if (pti[1] == DEFAULT_OPEN || isalnum((int) pti[1]))
 				{
 					return NULL;
 				}
@@ -1154,7 +1176,7 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 					case '7':
 					case '8':
 					case '9':
-						pti += isdigit(pti[2]) ? 3 : 2;
+						pti += isdigit((int) pti[2]) ? 3 : 2;
 						strcpy(pto, *pti == 0 ? "(.*)" : "(.*?)");
 						pto += strlen(pto);
 						break;
@@ -1209,7 +1231,7 @@ pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
 
 					case '?':
 						pti += 2;
-						strcpy(pto, *pti == 0 ? "(.?)" : "(.??)");
+						strcpy(pto, *pti == 0 ? "(.?)" : "(.?" "?)");
 						pto += strlen(pto);
 						break;
 
