@@ -96,6 +96,7 @@
 #define GET_ONE                          0 /* stop at spaces */
 #define GET_ALL                          1 /* stop at semicolon */
 #define GET_NST                          2 /* nest square brackets */
+#define GET_VBT                          4 /* ignore semicolon for verbatim mode */
 
 #define TEL_N                            0
 #define TEL_Y                            1
@@ -115,17 +116,16 @@
 
 #define HISTORY_FILE         ".tt_history"
 
-#define STRING_SIZE                  80000
-#define BUFFER_SIZE                  40000
+#define STRING_SIZE                  45000
+#define BUFFER_SIZE                  20000
 #define NUMBER_SIZE                    100
 #define LIST_SIZE                        2
 
-#define VERSION_NUM               "2.00.1"
+#define CLIENT_NAME              "TinTin++"
+#define CLIENT_VERSION           "2.00.8  "
 
 #define ESCAPE                          27
 
-#define TELNET_PORT                     23
-
 #define TIMER_POLL_INPUT                 0
 #define TIMER_POLL_SESSIONS              1
 #define TIMER_POLL_CHAT                  2
@@ -134,9 +134,11 @@
 #define TIMER_UPDATE_PACKETS             5
 #define TIMER_UPDATE_CHAT                6
 #define TIMER_UPDATE_TERMINAL            7
-#define TIMER_UPDATE_MEMORY              8
-#define TIMER_STALL_PROGRAM              9
-#define TIMER_CPU                       10
+#define TIMER_UPDATE_TIME                8
+#define TIMER_UPDATE_MEMORY              9
+#define TIMER_STALL_PROGRAM             10
+#define TIMER_CPU                       11
+
 
 #define PULSE_PER_SECOND                20
 
@@ -149,6 +151,7 @@
 #define PULSE_UPDATE_CHAT                2
 #define PULSE_UPDATE_TERMINAL            1
 #define PULSE_UPDATE_MEMORY              2
+#define PULSE_UPDATE_TIME               20
 
 /*
 	Index for lists used by tintin
@@ -290,9 +293,9 @@ enum operators
 #define TINTIN_FLAG_HISTORYBROWSE     (1 <<  2)
 #define TINTIN_FLAG_HISTORYSEARCH     (1 <<  3)
 #define TINTIN_FLAG_PROCESSINPUT      (1 <<  4)
-#define TINTIN_FLAG_USERCOMMAND       (1 <<  5)
+#define TINTIN_FLAG_USERCOMMAND       (1 <<  5) /* Unused */
 #define TINTIN_FLAG_INSERTINPUT       (1 <<  6)
-#define TINTIN_FLAG_VERBATIM          (1 <<  7)
+#define TINTIN_FLAG_VERBATIM          (1 <<  7) /* Unused */
 #define TINTIN_FLAG_TERMINATE         (1 <<  8)
 
 #define SES_FLAG_ECHOCOMMAND          (1 <<  1)
@@ -307,7 +310,7 @@ enum operators
 #define SES_FLAG_CONNECTED            (1 << 11)
 #define SES_FLAG_REPEATENTER          (1 << 12)
 #define SES_FLAG_VERBOSE              (1 << 13)
-#define SES_FLAG_REGEXP               (1 << 14) /* Unused */
+#define SES_FLAG_VERBOSELINE          (1 << 14)
 #define SES_FLAG_LOGLEVEL             (1 << 15)
 #define SES_FLAG_LOGPLAIN             (1 << 16)
 #define SES_FLAG_LOGHTML              (1 << 17)
@@ -319,7 +322,10 @@ enum operators
 #define SES_FLAG_SCROLLSTOP           (1 << 23)
 #define SES_FLAG_CONVERTMETA          (1 << 24)
 #define SES_FLAG_RUN                  (1 << 25)
-
+#define SES_FLAG_UTF8                 (1 << 26)
+#define SES_FLAG_BIG5                 (1 << 27)
+#define SES_FLAG_256COLOR             (1 << 28)
+#define SES_FLAG_IGNORELINE           (1 << 29)
 
 #define TELOPT_FLAG_SGA               (1 <<  0)
 #define TELOPT_FLAG_ECHO              (1 <<  1)
@@ -327,6 +333,8 @@ enum operators
 #define TELOPT_FLAG_PROMPT            (1 <<  3) 
 #define TELOPT_FLAG_DEBUG             (1 <<  4)
 #define TELOPT_FLAG_TSPEED            (1 <<  5)
+#define TELOPT_FLAG_TTYPE             (1 <<  6)
+#define TELOPT_FLAG_MTTS              (1 <<  7)
 
 #define LIST_FLAG_IGNORE              (1 <<  0)
 #define LIST_FLAG_MESSAGE             (1 <<  1)
@@ -350,12 +358,19 @@ enum operators
 #define ROOM_FLAG_LEAVE               (1 <<  2)
 #define ROOM_FLAG_VOID                (1 <<  3)
 #define ROOM_FLAG_STATIC              (1 <<  4)
+#define ROOM_FLAG_RIVER               (1 <<  5)
+#define ROOM_FLAG_PATH                (1 <<  6)
+
+#define EXIT_FLAG_HIDE                (1 <<  0)
+#define EXIT_FLAG_AVOID               (1 <<  1)
 
 #define MAP_FLAG_STATIC               (1 <<  0)
 #define MAP_FLAG_VTMAP                (1 <<  1)
 #define MAP_FLAG_VTGRAPHICS           (1 <<  2)
 #define MAP_FLAG_ASCIIGRAPHICS        (1 <<  3)
 #define MAP_FLAG_ASCIIVNUMS           (1 <<  4)
+#define MAP_FLAG_MUDFONT              (1 <<  5)
+#define MAP_FLAG_NOFOLLOW             (1 <<  6)
 
 #define MAP_EXIT_N                    (1 <<  0)
 #define MAP_EXIT_E                    (1 <<  1)
@@ -383,48 +398,48 @@ enum operators
 { \
 	if ((head) == NULL) \
 	{ \
-		(head)			   = (link); \
+		(head) = (link); \
 	} \
 	else \
 	{ \
-		(tail)->next		   = (link); \
+		(tail)->next = (link); \
 	} \
-	(link)->next			   = NULL; \
-	(link)->prev			   = (tail); \
+	(link)->next = NULL; \
+	(link)->prev = (tail); \
 	(tail)				    = (link); \
 }
 
 
-#define INSERT_LEFT(link, rght, head) \
+#define INSERT_LEFT(link, right, head) \
 { \
-	(link)->prev			   = (rght)->prev; \
-	(rght)->prev			   = (link); \
-	(link)->next			   = (rght); \
+	(link)->prev = (right)->prev; \
+	(right)->prev = (link); \
+	(link)->next = (right); \
 	\
 	if ((link)->prev) \
 	{ \
-		(link)->prev->next	  = (link); \
+		(link)->prev->next = (link); \
 	} \
 	else \
 	{ \
-		(head)			   = (link); \
+		(head) = (link); \
 	} \
 }
 
 
 #define INSERT_RIGHT(link, left, tail) \
 { \
-	(link)->next			   = (left)->next; \
-	(left)->next			   = (link); \
-	(link)->prev			   = (left); \
-	\
+	(link)->next = (left)->next; \
+	(left)->next = (link); \
+	(link)->prev = (left); \
+ \
 	if ((link)->next) \
 	{ \
-		(link)->next->prev	  = (link); \
+		(link)->next->prev = (link); \
 	} \
 	else \
 	{ \
-		(tail)			    = (link); \
+		(tail) = (link); \
 	} \
 }
 
@@ -472,32 +487,6 @@ enum operators
 	point = NULL; \
 }
 
-#define URANGE(a, b, c)           ((b) < (a) ? (a) : (b) > (c) ? (c) : (b))
-#define UMAX(a, b)                ((a) > (b) ? (a) : (b))
-#define UMIN(a, b)                ((a) < (b) ? (a) : (b))
-
-#define up(u)                     (u < 99 ? u++ : u)
-
-#define IS_SPLIT(ses)             ((ses)->rows != (ses)->bot_row || (ses)->top_row != 1)
-
-#define SCROLL(ses)               (((ses)->cur_row >= (ses)->top_row && (ses)->cur_row <= (ses)->bot_row) || (ses)->cur_row == (ses)->rows)
-
-#define VERBATIM                  (HAS_BIT(gtd->flags, TINTIN_FLAG_VERBATIM) && HAS_BIT(gtd->flags, TINTIN_FLAG_USERCOMMAND))
-
-#define DO_ARRAY(array) struct session *array (struct session *ses, struct listnode *list, char *arg)
-#define DO_CHAT(chat) void chat (char *left, char *right)
-#define DO_CLASS(class) struct session *class (struct session *ses, char *left, char *right)
-#define DO_COMMAND(command) struct session  *command (struct session *ses, char *arg)
-#define DO_CONFIG(config) struct session *config (struct session *ses, char *arg, int index)
-#define DO_MAP(map) void map (struct session *ses, char *arg, char *arg1, char *arg2)
-#define DO_PATH(path) void path (struct session *ses, char *arg)
-#define DO_LINE(line) void line (struct session *ses, char *arg)
-#define DO_CURSOR(cursor) void cursor (char *arg)
-#define DO_HISTORY(history) void history (struct session *ses, char *arg)
-#define DO_BUFFER(buffer) void buffer (struct session *ses, char *arg)
-
-
-
 /*
 	Bit operations
 */
@@ -508,6 +497,37 @@ enum operators
 #define TOG_BIT(bitvector, bit)   ((bitvector) ^= (bit))
 #define FFS_BIT(bitvector)        ((ffs(bitvector) - 1)) 
 
+/*
+	Generic
+*/
+
+#define URANGE(a, b, c)           ((b) < (a) ? (a) : (b) > (c) ? (c) : (b))
+#define UMAX(a, b)                ((a) > (b) ? (a) : (b))
+#define UMIN(a, b)                ((a) < (b) ? (a) : (b))
+
+#define up(u)                     (u < 99 ? u++ : u)
+
+#define IS_SPLIT(ses)             ((ses)->rows != (ses)->bot_row || (ses)->top_row != 1)
+
+#define SCROLL(ses)               ((ses)->cur_row == 0 || ((ses)->cur_row >= (ses)->top_row && (ses)->cur_row <= (ses)->bot_row) || (ses)->cur_row == (ses)->rows)
+
+#define VERBATIM(ses)             ((ses)->input_level == 0 && HAS_BIT((ses)->flags, SES_FLAG_VERBATIM))
+
+#define DO_ARRAY(array) struct session *array (struct session *ses, struct listnode *list, char *arg)
+#define DO_CHAT(chat) void chat (char *left, char *right)
+#define DO_CLASS(class) struct session *class (struct session *ses, char *left, char *right)
+#define DO_COMMAND(command) struct session  *command (struct session *ses, char *arg)
+#define DO_CONFIG(config) struct session *config (struct session *ses, char *arg, int index)
+#define DO_MAP(map) void map (struct session *ses, char *arg, char *arg1, char *arg2)
+#define DO_PATH(path) void path (struct session *ses, char *arg)
+#define DO_LINE(line) struct session *line (struct session *ses, char *arg)
+#define DO_CURSOR(cursor) void cursor (char *arg)
+#define DO_HISTORY(history) void history (struct session *ses, char *arg)
+#define DO_BUFFER(buffer) void buffer (struct session *ses, char *arg)
+
+
+
+
 
 /*
 	Compatibility
@@ -583,6 +603,8 @@ struct session
 	unsigned char         * read_buf;
 	int                     read_len;
 	int                     read_max;
+	int                     debug_level;
+	int                     input_level;
 	long long               connect_retry;
 	int                     connect_error;
 	char                    more_output[BUFFER_SIZE];
@@ -610,10 +632,12 @@ struct tintin_data
 	char                    input_tmp[BUFFER_SIZE];
 	char                    macro_buf[BUFFER_SIZE];
 	char                    paste_buf[BUFFER_SIZE];
+	int                     input_off;
 	int                     input_len;
 	int                     input_cur;
 	int                     input_pos;
-	int                     input_off;
+	int                     input_hid;
+	char                  * term;
 	long long               time;
 	long long               timer[TIMER_CPU][5];
 	long long               total_io_ticks;
@@ -667,6 +691,17 @@ struct link_data
 	char                 * str3;
 };
 
+struct grid_data
+{
+	struct grid_data     * next;
+	struct grid_data     * prev;
+	int                    vnum;
+	float                  length;
+	int                    x;
+	int                    y;
+	int                    z;
+};
+
 /*
 	Typedefs
 */
@@ -679,7 +714,7 @@ typedef struct session *COMMAND (struct session *ses, char *arg);
 typedef void            MAP     (struct session *ses, char *arg, char *arg1, char *arg2);
 typedef void            CURSOR  (char *arg);
 typedef void            PATH    (struct session *ses, char *arg);
-typedef void            LINE    (struct session *ses, char *arg);
+typedef struct session *LINE    (struct session *ses, char *arg);
 typedef void            HISTORY (struct session *ses, char *arg);
 typedef void            BUFFER  (struct session *ses, char *arg);
 
@@ -713,7 +748,6 @@ struct class_type
 struct color_type
 {
 	char                  * name;
-	char                  * number;
 	char                  * code;
 };
 
@@ -747,6 +781,12 @@ struct list_type
 	int                     flags;
 };
 
+struct substitution_type
+{
+	char                  * name;
+	int                     bitvector;
+};
+
 struct map_type
 {
 	char                  * name;
@@ -800,6 +840,12 @@ struct telopt_type
 	int                     flag;
 };
 
+struct term_type
+{
+	char                 * name;
+	int                    flag;
+};
+
 struct str_hash_data
 {
 	struct str_hash_data    * next;
@@ -823,15 +869,22 @@ struct map_data
 	FILE                  * logfile;
 	struct link_data      * undo_head;
 	struct link_data      * undo_tail;
-	char                  * room_color;
+	struct grid_data      * grid_head;
+	struct grid_data      * grid_tail;
 	char                  * exit_color;
+	char                  * path_color;
+	char                  * room_color;
 	int                     max_grid_x;
 	int                     max_grid_y;
 	int                     undo_size;
 	int                     size;
 	int                     flags;
 	int                     in_room;
+	int                     at_room;
 	int                     last_room;
+	short                   search_stamp;
+	short                   display_stamp;
+	short                   nofollow;
 	char                    legend[17][100];
 };
 
@@ -840,7 +893,10 @@ struct room_data
 	struct exit_data        * f_exit;
 	struct exit_data        * l_exit;
 	int                       vnum;
-	int                       size;
+	float                     length;
+	float                     weight;
+	short                     search_stamp;
+	short                     display_stamp;
 	int                       flags;
 	char                    * color;
 	char                    * name;
@@ -848,6 +904,8 @@ struct room_data
 	char                    * desc;
 	char                    * area;
 	char                    * note;
+	char                    * terrain;
+	char                    * data;
 };
 
 struct exit_data
@@ -856,8 +914,10 @@ struct exit_data
 	struct exit_data        * prev;
 	int                       vnum;
 	int                       dir;
+	int                       flags;
 	char                    * name;
 	char                    * cmd;
+	char                    * data;
 };
 
 
@@ -897,12 +957,13 @@ extern DO_CURSOR(cursor_buffer_home);
 extern DO_CURSOR(cursor_buffer_lock);
 extern DO_CURSOR(cursor_buffer_up);
 extern DO_CURSOR(cursor_check_line);
+extern DO_CURSOR(cursor_check_line_modified);
 extern DO_CURSOR(cursor_clear_left);
 extern DO_CURSOR(cursor_clear_line);
 extern DO_CURSOR(cursor_clear_right);
 extern DO_CURSOR(cursor_convert_meta);
 extern DO_CURSOR(cursor_delete);
-extern DO_CURSOR(cursor_delete_exit);
+extern DO_CURSOR(cursor_delete_or_exit);
 extern DO_CURSOR(cursor_delete_word_left);
 extern DO_CURSOR(cursor_delete_word_right);
 extern DO_CURSOR(cursor_echo_on);
@@ -910,7 +971,6 @@ extern DO_CURSOR(cursor_echo_off);
 extern DO_CURSOR(cursor_end);
 extern DO_CURSOR(cursor_enter);
 extern DO_CURSOR(cursor_exit);
-extern DO_CURSOR(cursor_fix_line);
 extern DO_CURSOR(cursor_history_find);
 extern DO_CURSOR(cursor_history_next);
 extern DO_CURSOR(cursor_history_prev);
@@ -925,6 +985,7 @@ extern DO_CURSOR(cursor_redraw_line);
 extern DO_CURSOR(cursor_right);
 extern DO_CURSOR(cursor_right_word);
 extern DO_CURSOR(cursor_suspend);
+extern DO_CURSOR(cursor_test);
 extern DO_CURSOR(cursor_tab_backward);
 extern DO_CURSOR(cursor_tab_forward);
 extern DO_CURSOR(cursor_auto_tab_backward);
@@ -943,7 +1004,7 @@ extern void read_line();
 extern void read_key(void);
 extern void convert_meta(char *input, char *output);
 extern void unconvert_meta(char *input, char *output);
-extern void echo_command(struct session *ses, char *line, int newline);
+extern void echo_command(struct session *ses, char *line);
 extern void input_printf(char *format, ...);
 extern void modified_input(void);
 
@@ -1048,7 +1109,9 @@ extern char *fix_file_name(char *name);
 #define __TERMINAL_H__
 
 extern void init_terminal(void);
-extern void reset_terminal(void);
+extern void restore_terminal(void);
+extern void refresh_terminal(void);
+
 extern void echo_on(struct session *ses);
 extern void echo_off(struct session *ses);
 extern void init_screen_size(struct session *ses);
@@ -1075,13 +1138,15 @@ extern void parse_class(struct session *ses, char *input, struct listnode *group
 
 extern DO_COMMAND(do_map);
 
+extern DO_MAP(map_at);
 extern DO_MAP(map_color);
 extern DO_MAP(map_create);
+extern DO_MAP(map_debug);
 extern DO_MAP(map_delete);
 extern DO_MAP(map_destroy);
 extern DO_MAP(map_dig);
-extern DO_MAP(map_exitcmd);
-extern DO_MAP(map_exitdir);
+extern DO_MAP(map_exit);
+extern DO_MAP(map_exitflag);
 extern DO_MAP(map_explore);
 extern DO_MAP(map_find);
 extern DO_MAP(map_flag);
@@ -1098,40 +1163,44 @@ extern DO_MAP(map_map);
 extern DO_MAP(map_move);
 extern DO_MAP(map_name);
 extern DO_MAP(map_read);
+extern DO_MAP(map_resize);
 extern DO_MAP(map_return);
 extern DO_MAP(map_roomflag);
 extern DO_MAP(map_run);
 extern DO_MAP(map_set);
 extern DO_MAP(map_travel);
 extern DO_MAP(map_undo);
+extern DO_MAP(map_uninsert);
 extern DO_MAP(map_unlink);
+extern DO_MAP(map_vnum);
 extern DO_MAP(map_write);
 
+
 extern void create_map(struct session *ses, char *arg);
 extern void delete_map(struct session *ses);
-extern int  create_room(struct session *ses, char *arg);
+extern int  create_room(struct session *ses, char *format, ...);
 extern void delete_room(struct session *ses, int room, int exits);
-extern void create_exit(struct session *ses, int room, char *arg);
+extern void create_exit(struct session *ses, int room, char *format, ...);
 extern void delete_exit(struct session *ses, int room, struct exit_data *exit);
 extern void create_legend(struct session *ses, char *arg);
-extern int match_room(struct session *ses, int room, char *arg1, char *arg2, char *arg3, char *arg4, char *arg5);
+extern int match_room(struct session *ses, int room, char *arg);
 extern int  find_room(struct session *ses, char *arg);
 extern void goto_room(struct session *ses, int room);
 extern struct exit_data *find_exit(struct session *ses, int room, char *arg);
 extern int  get_map_exit(struct session *ses, char *arg);
 extern int  get_map_exits(struct session *ses, int room);
-extern void create_map_grid(struct session *ses, int room, int x, int y);
-extern void build_map_grid(int room, int x, int y, int z);
-extern void follow_map(struct session *ses, char *argument);
+extern void displaygrid_build(struct session *ses, int room, int x, int y, int z);
+extern int  follow_map(struct session *ses, char *argument);
 extern void add_undo(struct session *ses, char *format, ...);
 extern void del_undo(struct session *ses, struct link_data *link);
+extern void del_gridnode(struct session *ses, struct grid_data *node);
 extern char *draw_room(struct session *ses, struct room_data *room, int line);
-extern void search_path(int room, int size, int dir);
-extern void fill_map(struct session *ses);
+extern int searchgrid_find(struct session *ses, int from, char *arg);
+extern int searchgrid_walk(struct session *ses, int from, int dest);
 extern void shortest_path(struct session *ses, int run, char *delay, char *arg);
 extern void explore_path(struct session *ses, int run, char *left, char *right);
 extern int find_coord(struct session *ses, char *arg);
-extern void search_coord_bf(int vnum);
+extern int spatialgrid_find(struct session *ses, int vnum, int x, int y, int z);
 extern void show_vtmap(struct session *ses);
 
 #endif
@@ -1141,7 +1210,6 @@ extern void show_vtmap(struct session *ses);
 
 extern DO_COMMAND(do_math);
 extern double get_number(struct session *ses, char *str);
-extern char *get_alnum(struct session *ses, char *str);
 extern void get_number_string(struct session *ses, char *str, char *result);
 extern double mathswitch(struct session *ses, char *left, char *right);
 extern void mathexp(struct session *ses, char *str, char *result);
@@ -1176,14 +1244,14 @@ extern void do_one_prompt(struct session *ses, char *prompt, int row);
 
 extern int substitute(struct session *ses, char *string, char *result, int flags);
 
-extern int match(struct session *ses, char *str, char *exp);
+extern int match(struct session *ses, char *str, char *exp, int flags);
 extern int find(struct session *ses, char *str, char *exp, int sub);
 DO_COMMAND(do_regexp);
 extern int regexp_compare(pcre *regex, char *str, char *exp, int option, int flag);
 extern int check_one_regexp(struct session *ses, struct listnode *node, char *line, char *original, int option);
-extern int tintin_regexp_check(char *exp);
-extern int tintin_regexp(pcre *pcre, char *str, char *exp, int option, int flag);
-extern pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option);
+extern int tintin_regexp_check(struct session *ses, char *exp);
+extern int tintin_regexp(struct session *ses, pcre *pcre, char *str, char *exp, int option, int flag);
+extern pcre *tintin_regexp_compile(struct session *ses, struct listnode *node, char *exp, int option);
 
 #endif
 
@@ -1215,6 +1283,8 @@ extern DO_CONFIG(config_loglevel);
 extern DO_CONFIG(config_colorpatch);
 extern DO_CONFIG(config_mccp);
 extern DO_CONFIG(config_autotab);
+extern DO_CONFIG(config_charset);
+extern DO_CONFIG(config_256color);
 
 #endif
 
@@ -1314,7 +1384,7 @@ extern void dump_full_stack(void);
 
 extern DO_COMMAND(do_event);
 extern DO_COMMAND(do_unevent);
-extern int check_all_events(struct session *ses, int args, int vars, char *fmt, ...);
+extern int check_all_events(struct session *ses, int flags, int args, int vars, char *fmt, ...);
 
 #endif
 
@@ -1402,9 +1472,11 @@ DO_HISTORY(history_write);
 #define __LINE_H__
 extern DO_COMMAND(do_line);
 extern DO_LINE(line_gag);
+extern DO_LINE(line_ignore);
 extern DO_LINE(line_log);
 extern DO_LINE(line_logverbatim);
-extern DO_LINE(line_skipgags);
+extern DO_LINE(line_substitute);
+extern DO_LINE(line_verbose);
 
 #endif
 
@@ -1482,6 +1554,7 @@ extern int get_nest_size(struct listroot *root, char *variable, char *result);
 extern struct listnode *get_nest_node(struct listroot *root, char *variable, char *result, int def);
 extern int get_nest_index(struct listroot *root, char *variable, char *result, int def);
 extern struct listnode *set_nest_node(struct listroot *root, char *arg1, char *format, ...);
+extern struct listnode *add_nest_node(struct listroot *root, char *arg1, char *format, ...);
 extern void show_nest_node(struct listnode *node, char *result, int initialize);
 extern void copy_nest_node(struct listroot *dst_root, struct listnode *dst, struct listnode *src);
 
@@ -1505,9 +1578,10 @@ extern int translate_telopts(struct session *ses, unsigned char *src, int cplen)
 extern int send_will_sga(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int send_do_eor(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int mark_prompt(struct session *ses, int cplen, unsigned char *cpsrc);
-extern int send_will_naws(struct session *ses, int cplen, unsigned char *cpsrc);
+extern int recv_do_naws(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int send_sb_naws(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int recv_sb_tspeed(struct session *ses, int cplen, unsigned char *cpsrc);
+extern int recv_dont_ttype(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int recv_sb_ttype(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int send_wont_status(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int send_dont_status(struct session *ses, int cplen, unsigned char *cpsrc);
@@ -1524,6 +1598,7 @@ extern int send_will_telopt(struct session *ses, int cplen, unsigned char *cpsrc
 extern int send_do_telopt(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int recv_sb_mssp(struct session *ses, int cplen, unsigned char *src);
 extern int recv_sb_msdp(struct session *ses, int cplen, unsigned char *src);
+extern int recv_sb_gmcp(struct session *ses, int cplen, unsigned char *src);
 extern int recv_sb_new_environ(struct session *ses, int cplen, unsigned char *src);
 extern int recv_sb_zmp(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int send_do_mssp(struct session *ses, int cplen, unsigned char *cpsrc);
@@ -1533,7 +1608,6 @@ extern int send_dont_mccp2(struct session *ses, int cplen, unsigned char *cpsrc)
 extern int init_mccp(struct session *ses, int cplen, unsigned char *cpsrc);
 extern void *zlib_alloc(void *opaque, unsigned int items, unsigned int size);
 extern void zlib_free(void *opaque, void *address);
-extern void init_telnet_session(struct session *ses);
 extern int skip_sb(struct session *ses, int cplen, unsigned char *cpsrc);
 extern int recv_sb(struct session *ses, int cplen, unsigned char *cpsrc);
 
@@ -1548,15 +1622,15 @@ extern struct session *parse_command(struct session *ses, char *input);
 extern int is_speedwalk(struct session *ses, char *input);
 extern void process_speedwalk(struct session *ses, char *input);
 extern struct session *parse_tintin_command(struct session *ses, char *input);
-extern char *get_arg_all(char *string, char *result);
-extern char *get_arg_in_braces(char *string, char *result, int flag);
+extern char *get_arg_all(struct session *ses, char *string, char *result, int verbatim);
+extern char *get_arg_in_braces(struct session *ses, char *string, char *result, int flag);
 extern char *sub_arg_in_braces(struct session *ses, char *string, char *result, int flag, int sub);
-extern char *get_arg_with_spaces(char *string, char *result, int flag);
-extern char *get_arg_stop_spaces(char *string, char *result, int flag);
+extern char *get_arg_with_spaces(struct session *ses, char *string, char *result, int flag);
+extern char *get_arg_stop_spaces(struct session *ses, char *string, char *result, int flag);
 extern char *space_out(char *string);
-extern char *get_arg_to_brackets(char *string, char *result);
-extern char *get_arg_at_brackets(char *string, char *result);
-extern char *get_arg_in_brackets(char *string, char *result);
+extern char *get_arg_to_brackets(struct session *ses, char *string, char *result);
+extern char *get_arg_at_brackets(struct session *ses, char *string, char *result);
+extern char *get_arg_in_brackets(struct session *ses, char *string, char *result);
 extern void write_mud(struct session *ses, char *command, int flags);
 extern void do_one_line(char *line, struct session *ses);
 
@@ -1570,10 +1644,11 @@ extern DO_PATH(path_del);
 extern DO_PATH(path_end);
 extern DO_PATH(path_ins);
 extern DO_PATH(path_load);
-extern DO_PATH(path_map);
 extern DO_PATH(path_new);
 extern DO_PATH(path_run);
 extern DO_PATH(path_save);
+extern DO_PATH(path_show);
+extern DO_PATH(path_unzip);
 extern DO_PATH(path_walk);
 extern DO_PATH(path_zip);
 
@@ -1642,6 +1717,7 @@ extern struct config_type config_table[];
 extern struct cursor_type cursor_table[];
 extern struct event_type event_table[];
 extern struct list_type list_table[LIST_MAX];
+extern struct substitution_type substitution_table[];
 extern struct map_type map_table[];
 extern struct timer_type timer_table[];
 extern struct path_type path_table[];
@@ -1649,6 +1725,7 @@ extern struct line_type line_table[];
 extern struct history_type history_table[];
 extern struct buffer_type buffer_table[];
 extern struct telopt_type telopt_table[];
+extern struct term_type term_table[];
 
 #endif
 
@@ -1657,6 +1734,7 @@ extern struct telopt_type telopt_table[];
 
 extern void printline(struct session *ses, char *str, int isaprompt);
 int word_wrap(struct session *ses, char *textin, char *textout, int display);
+int word_wrap_split(struct session *ses, char *textin, char *textout, int skip, int keep);
 
 #endif
 
@@ -1691,7 +1769,7 @@ extern void packet_update(void);
 extern void chat_update(void);
 extern void terminal_update(void);
 extern void memory_update(void);
-
+extern void time_update(void);
 #endif
 
 #ifndef __UTILS_H__
@@ -1706,7 +1784,7 @@ extern long long utime(void);
 extern char *capitalize(char *str);
 extern char *ntos(long long number);
 extern char *indent(int cnt);
-extern void cat_sprintf(char *dest, char *fmt, ...);
+extern int cat_sprintf(char *dest, char *fmt, ...);
 extern void ins_sprintf(char *dest, char *fmt, ...);
 extern int str_suffix(char *str1, char *str2);
 extern void syserr(char *msg);
@@ -1759,7 +1837,6 @@ extern DO_COMMAND(do_replace);
 
 extern void save_pos(struct session *ses);
 extern void restore_pos(struct session *ses);
-extern void restore_cursor(struct session *ses);
 extern void goto_rowcol(struct session *ses, int row, int col);
 extern void erase_screen(struct session *ses);
 extern void erase_toeol(void);
@@ -1773,7 +1850,7 @@ extern void strip_vt102_codes(char *str, char *buf);
 extern void strip_vt102_codes_non_graph(char *str, char *buf);
 extern void strip_non_vt102_codes(char *str, char *buf);
 extern void get_color_codes(char *old, char *str, char *buf);
-extern int strip_vt102_strlen(char *str);
+extern int strip_vt102_strlen(struct session *ses, char *str);
 extern int strip_color_strlen(char *str);
 extern int interpret_vt102_codes(struct session *ses, char *str, int real);
 
