@@ -27,8 +27,6 @@
 
 #include "tintin.h"
 
-int debug_level;
-
 struct scriptdata
 {
 	long long              min;
@@ -74,7 +72,7 @@ void debugtoken(struct session *ses, struct scriptnode *token)
 {
 	push_call("debugtoken(%p,%p,%d)",ses,token,token->type);
 
-	if (debug_level)
+	if (ses->debug_level)
 	{
 		switch (token->type)
 		{
@@ -160,9 +158,9 @@ char *addlooptoken(struct scriptroot *root, int lvl, int opr, int cmd, char *str
 
 	data = (struct scriptdata *) calloc(1, sizeof(struct scriptdata));
 
-	str = get_arg_in_braces(str, min, FALSE);
-	str = get_arg_in_braces(str, max, FALSE);
-	str = get_arg_in_braces(str, var, FALSE);
+	str = get_arg_in_braces(root->ses, str, min, FALSE);
+	str = get_arg_in_braces(root->ses, str, max, FALSE);
+	str = get_arg_in_braces(root->ses, str, var, FALSE);
 
 	data->cpy = refstring(NULL, "{%s} {%s} {%s}", min, max, var);
 
@@ -175,14 +173,35 @@ char *addlooptoken(struct scriptroot *root, int lvl, int opr, int cmd, char *str
 	return str;
 }
 
+char *addswitchtoken(struct scriptroot *root, int lvl, int opr, int cmd, char *str)
+{
+	struct scriptdata *data;
+
+	char arg[BUFFER_SIZE];
+
+	data = (struct scriptdata *) calloc(1, sizeof(struct scriptdata));
+
+	str = get_arg_in_braces(root->ses, str, arg, FALSE);
+
+	data->cpy = strdup("");
+
+	addtoken(root, lvl, opr, cmd, arg);
+
+	data->str = strdup("");
+
+	root->prev->data = data;
+
+	return str;
+}
+
 void resetlooptoken(struct session *ses, struct scriptnode *token)
 {
 	char *str, min[BUFFER_SIZE], max[BUFFER_SIZE];
 
 	str = token->data->cpy;
 
-	str = get_arg_in_braces(str, min, FALSE);
-	str = get_arg_in_braces(str, max, FALSE);
+	str = get_arg_in_braces(ses, str, min, FALSE);
+	str = get_arg_in_braces(ses, str, max, FALSE);
 
 	token->data->min = (int) get_number(ses, min);
 	token->data->max = (int) get_number(ses, max);
@@ -202,8 +221,8 @@ char *addforeachtoken(struct scriptroot *root, int lvl, int opr, int cmd, char *
 
 	char arg[BUFFER_SIZE], var[BUFFER_SIZE];
 
-	str = get_arg_in_braces(str, arg, FALSE);
-	str = get_arg_in_braces(str, var, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg, FALSE);
+	str = get_arg_in_braces(root->ses, str, var, FALSE);
 
 	addtoken(root, lvl, opr, cmd, var);
 
@@ -247,11 +266,20 @@ void handlereturntoken(struct session *ses, struct scriptnode *token)
 	set_nest_node(ses->list[LIST_VARIABLE], "result", "%s", arg);
 }
 
-char *get_arg_foreach(struct scriptnode *token)
+void handleswitchtoken(struct session *ses, struct scriptnode *token)
+{
+	char arg[BUFFER_SIZE];
+
+	mathexp(ses, token->str, arg);
+
+	RESTRING(token->data->str, arg);
+}
+
+char *get_arg_foreach(struct scriptroot *root, struct scriptnode *token)
 {
 	static char buf[BUFFER_SIZE];
 
-	token->data->arg = get_arg_in_braces(token->data->arg, buf, TRUE);
+	token->data->arg = get_arg_in_braces(root->ses, token->data->arg, buf, TRUE);
 
 	if (*token->data->arg == COMMAND_SEPARATOR)
 	{
@@ -266,8 +294,8 @@ char *addparsetoken(struct scriptroot *root, int lvl, int opr, int cmd, char *st
 
 	char arg[BUFFER_SIZE], var[BUFFER_SIZE];
 
-	str = get_arg_in_braces(str, arg, FALSE);
-	str = get_arg_in_braces(str, var, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg, FALSE);
+	str = get_arg_in_braces(root->ses, str, var, FALSE);
 
 	addtoken(root, lvl, opr, cmd, var);
 
@@ -303,11 +331,33 @@ void breakparsetoken(struct scriptnode *token)
 	token->data->arg = token->data->str;
 }
 
-char *get_arg_parse(struct scriptnode *token)
+char *get_arg_parse(struct session *ses, struct scriptnode *token)
 {
-	static char buf[2];
+	static char buf[5];
 
-	buf[0] = *token->data->arg++;
+	if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && token->data->arg[0] & 128 && token->data->arg[1] != 0)
+	{
+		token->data->arg += sprintf(buf, "%c%c", token->data->arg[0], token->data->arg[1]);
+	}
+	else if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (token->data->arg[0] & 192) == 192 && token->data->arg[1] != 0)
+	{
+		if ((token->data->arg[0] & 240) == 240 && token->data->arg[2] != 0 && token->data->arg[3] != 0)
+		{
+			token->data->arg += sprintf(buf, "%c%c%c%c", token->data->arg[0], token->data->arg[1], token->data->arg[2], token->data->arg[3]);
+		}
+		else if ((token->data->arg[0] & 224) == 224 && token->data->arg[2] != 0)
+		{
+			token->data->arg += sprintf(buf, "%c%c%c", token->data->arg[0], token->data->arg[1], token->data->arg[2]);
+		}
+		else
+		{
+			token->data->arg += sprintf(buf, "%c%c", token->data->arg[0], token->data->arg[1]);
+		}
+	}
+	else
+	{
+		token->data->arg += sprintf(buf, "%c", token->data->arg[0]);
+	}
 
 	return buf;
 }
@@ -318,9 +368,9 @@ char *addregextoken(struct scriptroot *root, int lvl, int type, int cmd, char *s
 
 	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], arg3[BUFFER_SIZE];
 
-	str = get_arg_in_braces(str, arg1, FALSE);
-	str = get_arg_in_braces(str, arg2, FALSE);
-	str = get_arg_in_braces(str, arg3, TRUE);
+	str = get_arg_in_braces(root->ses, str, arg1, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg2, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg3, TRUE);
 
 	addtoken(root, lvl, type, cmd, arg1);
 
@@ -355,6 +405,7 @@ void deltoken(struct scriptroot *root, struct scriptnode *token)
 		case TOKEN_TYPE_LOOP:
 		case TOKEN_TYPE_FOREACH:
 		case TOKEN_TYPE_PARSE:
+		case TOKEN_TYPE_SWITCH:
 			free(token->data->cpy);
 			free(token->data->str);
 			free(token->data);
@@ -411,26 +462,26 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 
 	while (*str)
 	{
-		if (!VERBATIM)
+		if (!VERBATIM(root->ses))
 		{
 			str = space_out(str);
 		}
 
 		if (*str != gtd->tintin_char)
 		{
-			str = get_arg_all(str, line);
+			str = get_arg_all(root->ses, str, line, VERBATIM(root->ses));
 
 			addtoken(root, lvl, TOKEN_TYPE_STRING, -1, line);
 		}
 		else
 		{
-			arg = get_arg_stop_spaces(str, line, 0);
+			arg = get_arg_stop_spaces(root->ses, str, line, 0);
 
 			cmd = find_command(line+1);
 
 			if (cmd == -1)
 			{
-				str = get_arg_all(str, line);
+				str = get_arg_all(root->ses, str, line, FALSE);
 				addtoken(root, lvl, TOKEN_TYPE_SESSION, -1, line+1);
 			}
 			else
@@ -438,29 +489,29 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 				switch (command_table[cmd].type)
 				{
 					case TOKEN_TYPE_BREAK:
-						str = get_arg_with_spaces(arg, line, 1);
+						str = get_arg_with_spaces(root->ses, arg, line, 1);
 						addtoken(root, lvl, TOKEN_TYPE_BREAK, cmd, line);
 						break;
 
 					case TOKEN_TYPE_CASE:
-						str = get_arg_in_braces(arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, FALSE);
 						addtoken(root, lvl++, TOKEN_TYPE_CASE, cmd, line);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endcase");
 						break;
 
 					case TOKEN_TYPE_CONTINUE:
-						str = get_arg_with_spaces(arg, line, 1);
+						str = get_arg_with_spaces(root->ses, arg, line, 1);
 						addtoken(root, lvl, TOKEN_TYPE_CONTINUE, cmd, line);
 						break;
 
 					case TOKEN_TYPE_DEFAULT:
 						addtoken(root, lvl++, TOKEN_TYPE_DEFAULT, cmd, "");
 
-						str = get_arg_in_braces(arg, line, TRUE);
+						str = get_arg_in_braces(root->ses, arg, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "enddefault");
@@ -469,17 +520,17 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_ELSE:
 						addtoken(root, lvl++, TOKEN_TYPE_ELSE, cmd, "else");
 
-						str = get_arg_in_braces(arg, line, TRUE);
+						str = get_arg_in_braces(root->ses, arg, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endelse");
 						break;
 
 					case TOKEN_TYPE_ELSEIF:
-						str = get_arg_in_braces(arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, FALSE);
 						addtoken(root, lvl++, TOKEN_TYPE_ELSEIF, cmd, line);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endif");
@@ -488,17 +539,17 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_FOREACH:
 						str = addforeachtoken(root, lvl++, TOKEN_TYPE_FOREACH, cmd, arg);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endforeach");
 						break;
 
 					case TOKEN_TYPE_IF:
-						str = get_arg_in_braces(arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, FALSE);
 						addtoken(root, lvl++, TOKEN_TYPE_IF, cmd, line);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endif");
@@ -507,7 +558,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 						{
 							addtoken(root, lvl++, TOKEN_TYPE_ELSE, -1, "else");
 
-							str = get_arg_in_braces(str, line, TRUE);
+							str = get_arg_in_braces(root->ses, str, line, TRUE);
 							tokenize_script(root, lvl--, line);
 
 							addtoken(root, lvl, TOKEN_TYPE_END, -1, "endif");
@@ -517,7 +568,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_LOOP:
 						str = addlooptoken(root, lvl++, TOKEN_TYPE_LOOP, cmd, arg);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endloop");
@@ -526,22 +577,22 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_PARSE:
 						str = addparsetoken(root, lvl++, TOKEN_TYPE_PARSE, cmd, arg);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endparse");
 						break;
 
 					case TOKEN_TYPE_REGEX:
-						str = addregextoken(root, lvl++, TOKEN_TYPE_REGEX, cmd, arg);
+						str = addregextoken(root, lvl, TOKEN_TYPE_REGEX, cmd, arg);
 
-						addtoken(root, --lvl, TOKEN_TYPE_END, -1, "endregex");
+//						addtoken(root, --lvl, TOKEN_TYPE_END, -1, "endregex");
 
 						if (*str && *str != COMMAND_SEPARATOR)
 						{
 							addtoken(root, lvl++, TOKEN_TYPE_ELSE, -1, "else");
 
-							str = get_arg_in_braces(str, line, TRUE);
+							str = get_arg_in_braces(root->ses, str, line, TRUE);
 							tokenize_script(root, lvl--, line);
 
 							addtoken(root, lvl, TOKEN_TYPE_END, -1, "endregex");
@@ -549,32 +600,31 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 						break;
 
 					case TOKEN_TYPE_RETURN:
-						str = get_arg_with_spaces(arg, line, 1);
+						str = get_arg_with_spaces(root->ses, arg, line, 1);
 						addtoken(root, lvl, TOKEN_TYPE_RETURN, cmd, line);
 						break;
 
 					case TOKEN_TYPE_SWITCH:
-						str = get_arg_in_braces(arg, line, FALSE);
-						addtoken(root, lvl++, TOKEN_TYPE_SWITCH, cmd, line);
+						str = addswitchtoken(root, lvl++, TOKEN_TYPE_SWITCH, cmd, arg);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "end");
 						break;
 
 					case TOKEN_TYPE_WHILE:
-						str = get_arg_in_braces(arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, FALSE);
 						addtoken(root, lvl++, TOKEN_TYPE_WHILE, cmd, line);
 
-						str = get_arg_in_braces(str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, TRUE);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endwhile");
 						break;
 
 					default:
-						str = get_arg_with_spaces(arg, line, 1);
+						str = get_arg_with_spaces(root->ses, arg, line, 1);
 						addtoken(root, lvl, TOKEN_TYPE_COMMAND, cmd, line);
 						break;
 				}
@@ -654,7 +704,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 				continue;
 
 			case TOKEN_TYPE_CASE:
-				if (mathswitch(root->ses, shift->str, token->str))
+				if (mathswitch(root->ses, shift->data->str, token->str))
 				{
 					token = token->next;
 
@@ -755,7 +805,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 				}
 				else
 				{
-					set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%s", get_arg_foreach(token));
+					set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%s", get_arg_foreach(root, token));
 
 					if (*token->data->arg == 0)
 					{
@@ -807,9 +857,23 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 				if (*token->data->arg == 0)
 				{
 					resetparsetoken(root->ses, token);
+
+					if (*token->data->arg == 0)
+					{
+						token->type++;
+
+						do
+						{
+							token = token->next;
+						}
+						while (token && token->lvl > lvl);
+
+						continue;
+					}
+
 				}
 
-				set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%s", get_arg_parse(token));
+				set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%s", get_arg_parse(root->ses, token));
 
 				if (*token->data->arg == 0)
 				{
@@ -858,7 +922,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 				break;
 
 			case TOKEN_TYPE_SWITCH:
-				utime();
+				handleswitchtoken(root->ses, token);
 
 				token = parse_script(root, lvl + 1, token->next, token);
 				continue;
@@ -947,7 +1011,7 @@ char *write_script(struct session *ses, struct scriptroot *root)
 				break;
 
 			case TOKEN_TYPE_REGEX:
-				cat_sprintf(buf, "%s%c%s {%s} {%s} {%s}\n", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, token->regex->str, token->regex->bod);
+				cat_sprintf(buf, "%s%c%s {%s} {%s} {%s}", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, token->regex->str, token->regex->bod);
 				break;
 
 			case TOKEN_TYPE_END:
@@ -984,21 +1048,25 @@ char *write_script(struct session *ses, struct scriptroot *root)
 struct session *script_driver(struct session *ses, int list, char *str)
 {
 	struct scriptroot *root;
-	int level;
+	struct session *cur_ses;
+	int dlevel, ilevel;
 
 	root = (struct scriptroot *) calloc(1, sizeof(struct scriptroot));
 
-	root->ses = ses;
+	root->ses = cur_ses = ses;
+
+	dlevel = (list >= 0) ? HAS_BIT(root->ses->list[list]->flags, LIST_FLAG_DEBUG) : 0;
+	ilevel = (list >= 0) ? 1 : 0;
+
+	cur_ses->debug_level += dlevel;
+	cur_ses->input_level += ilevel;
 
 	tokenize_script(root, 0, str);
 
-	level = list != -1 ? HAS_BIT(ses->list[list]->flags, LIST_FLAG_DEBUG) : 0;
-
-	debug_level += level;
-
 	ses = (struct session *) parse_script(root, 0, root->next, root->prev);
 
-	debug_level -= level;
+	cur_ses->debug_level -= dlevel;
+	cur_ses->input_level -= ilevel;
 
 	while (root->prev)
 	{
